import fiber
import pickle
import time
import pytest
import docker
import collections
import multiprocessing as mp
from fiber.queues import Pipe, SimpleQueue
from fiber.socket import Socket


@pytest.fixture(scope="module")
def client():
    return docker.from_env()


def write_pipe(pipe, msg):
    pipe.send(msg)


def put_queue(q, data):
    print("put_queue started")
    if type(data) is list:
        for d in data:
            q.put(d)
    else:
        q.put(data)
    # make sure all data are flushed out
    time.sleep(1)
    print("put_queue done")


def get_queue(q_in, q_out, n):
    print("get_queue started", q_in, q_in.reader, q_in.writer)
    for i in range(n):
        d = q_in.get()
        q_out.put(d)
    time.sleep(1)
    print("get_queue finished")


def worker(q_in, q_out, ident):
    data = None
    while True:
        data = q_in.get()
        if data == "quit":
            break
        q_out.put(ident)


def pipe_worker(conn):
    d = conn.recv()
    print("[pipe_worker] got command", d)
    conn.send(b"ack")
    print("[pipe_worker] ack")


class TestZconnection():

    def test_zconn_with_mp(self):
        mp.Process()
        pass

def pipe_mp_worker(writer):
    import multiprocessing as mp
    ctx = mp.get_context('spawn')
    p = ctx.Process(target=write_pipe, args=(writer, b"hello"), name="WRITE_PIPE")
    p.start()

class TestQueue():

    @pytest.fixture(autouse=True)
    def check_leak(self, request, client):
        assert fiber.active_children() == []
        if fiber.config.default_backend != "docker":
            yield
            return
        pre_workers = client.containers.list()
        yield
        post_workers = client.containers.list()
        assert pre_workers == post_workers

    def test_pipe(self):
        reader, writer = Pipe()

        writer.send(b"hello")
        data = reader.recv()
        assert data == b"hello"

    def test_sharing_pipe_with_multiprocessing(self):
        reader, writer = Pipe()
        # The contaxt is 'spawn' because fork based mp won't trigger ZConnection's
        # initialization.
        ctx = mp.get_context('spawn')
        p = ctx.Process(target=write_pipe, args=(writer, b"hello"))
        p.start()
        msg = reader.recv()
        p.join()
        assert msg == b"hello"

    def test_subprocess_with_pipe(self):
        reader, writer = Pipe()
        p = fiber.Process(target=write_pipe, args=(writer, b"fiber pipe"))
        p.start()
        msg = reader.recv()
        p.join()
        assert msg == b"fiber pipe"

    def test_pipe_duplex1(self):
        conn1, conn2 = Pipe(duplex=True)

        conn1.send(b"hello")
        data = conn2.recv()
        assert data == b"hello"

    def test_pipe_duplex2(self):
        conn1, conn2 = Pipe(duplex=True)

        conn2.send(b"hi")
        data = conn1.recv()
        assert data == b"hi"

    def test_pipe_duplex_over_fiber_process(self):
        conn1, conn2 = Pipe(duplex=True)
        p = fiber.Process(target=pipe_worker, args=(conn2,))
        p.start()
        conn1.send(b"hello")
        data = conn1.recv()
        p.join()
        assert data == b"ack"

    def test_sharing_pipe_with_fiber_and_multiprocessing(self):
        reader, writer = Pipe()

        p = fiber.Process(target=pipe_mp_worker, args=(writer,))
        p.start()
        msg = reader.recv()
        p.join()
        assert msg == b"hello"

    def test_simple_queue(self):
        q = SimpleQueue()
        ctx = mp.get_context('spawn')
        p = ctx.Process(target=put_queue, args=(q, 10))
        p.start()
        p.join()
        data = q.get()
        assert data == 10

    def test_simple_queue_fiber(self):
        q = SimpleQueue()
        p = fiber.Process(target=put_queue, args=(q, 10))
        p.start()
        p.join()
        data = q.get()
        assert data == 10

    def test_simple_queue_fiber2(self):
        q = SimpleQueue()
        n = 3
        procs = []
        try:
            for i in range(n):
                procs.append(fiber.Process(target=put_queue, args=(q, 10)))
            for p in procs:
                p.start()
            for p in procs:
                p.join()
            for p in procs:
                data = q.get()
                assert data == 10
        finally:
            for p in procs:
                p.join()

    def test_simple_queue_fiber_multi(self):
        # Read and write multiple objects
        n = 10
        q = SimpleQueue()
        p = fiber.Process(target=put_queue, args=(q, [i for i in range(n)]))
        p.start()
        p.join()
        for i in range(n):
            data = q.get()
            assert data == i

    def test_simple_queue_read_write_from_different_proc(self):
        # One process writew, another process reads and then writes to q_out.
        # Then master process reads from q_out.
        n = 10
        q = SimpleQueue()
        q_out = SimpleQueue()
        p1 = fiber.Process(target=put_queue, args=(q, [i for i in range(n)]))
        p2 = fiber.Process(target=get_queue, args=(q, q_out, n))
        p1.start()
        p2.start()
        for i in range(n):
            data = q_out.get()
            assert data == i
        p1.join()
        p2.join()

    @pytest.mark.skip(reason="both reader and writer are initialized to LazyZConnection")
    def test_queue_reader_is_none_by_default(self):
        # set reader to None so that de-serizlied Queue doesn't connect to
        # reader automatically. Fiber socket will faily queue all the messages
        # to all the readers. We want to prevent this behavior.

        # This is only needed by SimpleQueuePush
        q = fiber.queues.SimpleQueuePush()
        q.put(1)
        q.get()
        d = pickle.dumps(q)
        q2 = pickle.loads(d)
        assert q2.reader is None
        assert q2.writer is not None

    def test_queue_balance(self):
        # We only test SimpleQueuePush because SimpleQueuePull doesn't gurantee
        # balance.
        inqueue = fiber.queues.SimpleQueuePush()
        outqueue = fiber.queues.SimpleQueuePush()
        num_workers = 4
        multiplier = 600
        workers = []
        results = []
        for i in range(num_workers):
            print("create worker", i)
            p = fiber.Process(target=worker, args=(inqueue, outqueue, i), daemon=True)
            workers.append(p)
        for i in range(num_workers):
            workers[i].start()

        # wait for all workers to connect
        time.sleep(1)
        for i in range(num_workers * multiplier):
            inqueue.put("work")
        for i in range(num_workers * multiplier):
            results.append(outqueue.get())
        stats = collections.Counter(results)
        total = num_workers * multiplier
        # send singals to all workers
        for i in range(num_workers * multiplier):
            inqueue.put("quit")
        for i in range(num_workers):
            workers[i].join()
        for i in range(num_workers):
            #print("{}: {} {:.2f}".format(i, stats[i], stats[i] / float(total)))
            # data should be fairly queued
            assert stats[i] == 600



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Kubernetes backend. Helps Fiber to launch jobs on Kubernetes.
"""

import logging
import multiprocessing as mp
import socket
import time
import uuid

import psutil
from kubernetes import client, config
from kubernetes.client.rest import ApiException

import fiber
import fiber.core as core
import fiber.config as fiber_config
from fiber.core import ProcessStatus
from fiber.util import find_ip_by_net_interface, find_listen_address


logger = logging.getLogger("fiber")


PHASE_STATUS_MAP = {
    "Pending": ProcessStatus.INITIAL,
    "Running": ProcessStatus.STARTED,
    "Succeeded": ProcessStatus.STOPPED,
    "Failed": ProcessStatus.STOPPED,
    "Unknown": ProcessStatus.STOPPED,
}


class Backend(core.Backend):
    name = "kubernetes"

    def __init__(self, incluster=True):
        if incluster:
            config.load_incluster_config()
        else:
            config.load_kube_config()
        self.config = client.Configuration()
        self.batch_api = client.BatchV1Api(client.ApiClient(self.config))
        self.core_api = client.CoreV1Api(client.ApiClient(self.config))
        self.default_namespace = fiber_config.kubernetes_namespace

        if incluster:
            podname = socket.gethostname()
            logger.debug(f'searching {podname} in {self.default_namespace}')
            pod = self.core_api.read_namespaced_pod(podname,
                                                    self.default_namespace)
            # Current model assume that Fiber only lauches 1 container per pod
            self.current_image = pod.spec.containers[0].image
            self.volumes = pod.spec.volumes
            self.mounts = pod.spec.containers[0].volume_mounts

            #if pod.spec.volumes:
            #    self.current_mount = pod.spec.volumes[0].persistent_volume_claim.claim_name
            #else:
            #    self.current_mount = None
        else:
            self.current_image = None
            self.mounts = None
            self.volumes = None

    def _get_resource_requirements(self, job_spec):
        #requests = {}
        limits = {}

        if job_spec.cpu:
            limits["cpu"] = str(job_spec.cpu)
        else:
            # by default, allocate 1 cpu
            limits["cpu"] = str(1)

        if job_spec.mem:
            limits["memory"] = str(job_spec.mem) + "Mi"

        if job_spec.gpu:
            limits["nvidia.com/gpu"] = str(job_spec.gpu)

        if limits != {}:
            rr = client.V1ResourceRequirements()
            rr.limits = limits
            return rr

        return None


    def create_job(self, job_spec):
        logger.debug("[k8s]create_job: %s", job_spec)
        body = client.V1Pod()
        name = "{}-{}".format(
            job_spec.name.replace("_", "-").lower(),
            str(uuid.uuid4())[:8]
        )
        body.metadata = client.V1ObjectMeta(
            namespace=self.default_namespace, name=name
        )

        # set environment varialbes
        # TODO(jiale) add environment variables

        image = job_spec.image if job_spec.image else self.current_image

        container = client.V1Container(
            name=name, image=image, command=job_spec.command, env=[],
            stdin=True, tty=True,
        )

        rr = self._get_resource_requirements(job_spec)
        if rr:
            logger.debug(
                "[k8s]create_job, container resource requirements: %s",
                job_spec
            )
            container.resources = rr

        body.spec = client.V1PodSpec(
            containers=[container],
            restart_policy="Never"
        )

        # propagate mount points to new containers if necesary
        if job_spec.volumes:
            volumes = []
            volume_mounts = []

            for pd_name, mount_info in job_spec.volumes.items():
            #volume_name = job_spec.volume if job_spec.volume else self.current_mount
                pvc = client.V1PersistentVolumeClaimVolumeSource(
                    claim_name=pd_name
                )
                volume = client.V1Volume(
                    persistent_volume_claim=pvc,
                    name="volume-" + pd_name,
                )
                volumes.append(volume)
                mount = client.V1VolumeMount(
                    mount_path=mount_info["bind"],
                    name=volume.name,
                )
                if mount_info["mode"] == "r":
                    mount.read_only = True
                volume_mounts.append(mount)
            body.spec.volumes = volumes
            container.volume_mounts = volume_mounts
        elif self.mounts:
            body.spec.volumes = self.volumes
            container.volume_mounts = self.mounts

        logger.debug("[k8s]calling create_namespaced_pod: %s", body.metadata.name)
        try:
            v1pod = self.core_api.create_namespaced_pod(
                self.default_namespace, body
            )
        except ApiException as e:
            raise e

        return core.Job(v1pod, v1pod.metadata.uid)

    def get_job_status(self, job):
        v1pod = job.data
        name = v1pod.metadata.name
        namespace = v1pod.metadata.namespace

        try:
            v1pod = self.core_api.read_namespaced_pod_status(name, namespace)
            logger.debug("get_job_status: v1pod: %s", v1pod.status.phase)
        except ApiException as e:
            logger.debug(
                "[k8s] Exception when calling CoreApi->"
                "read_namespaced_pod_status: %s",
                str(e),
            )
            return ProcessStatus.STOPPED

        container_statuses = v1pod.status.container_statuses
        if container_statuses:
            # container_statuses[0].state.terminated is updated faster than pod_status.phase
            if container_statuses[0].state.terminated:
                return ProcessStatus.STOPPED
        pod_status = v1pod.status
        return PHASE_STATUS_MAP[pod_status.phase]

    def get_job_logs(self, job):
        v1job = job.data
        name = v1job.metadata.name
        namespace = v1job.metadata.namespace

        try:
            logs = self.batch_api.read_namespaced_pod_log(name, namespace)
        except ApiException as e:
            logger.debug(
                "[k8s] Exception when calling BatchV1Api->"
                "read_namespaced_job_status: %s",
                str(e),
            )
            raise e

        return logs

    def wait_for_job(self, job, timeout):
        logger.debug("[k8s]wait_for_job timeout=%s", timeout)

        total = 0
        while self.get_job_status(job) != ProcessStatus.STOPPED:
            time.sleep(1)
            total += 1
            if timeout is None:
                # timeout is None means wait until process exits
                continue
            if total > timeout:
                logger.debug("[k8s]wait_for_job done(timeout)")
                return None

        v1pod = job.data
        name = v1pod.metadata.name
        namespace = v1pod.metadata.namespace
        try:
            v1pod = self.core_api.read_namespaced_pod_status(name, namespace)
        except ApiException as e:
            logger.debug(
                "[k8s] Exception when calling CoreApi->"
                "read_namespaced_pod_status: %s",
                str(e),
            )
            # job has been deleted
            return -1

        pod_status = v1pod.status

        terminated = pod_status.container_statuses[0].state.terminated
        if terminated is None:
            logger.debug("[k8s]wait_for_job done: container is not terminated")
            return None

        logger.debug("[k8s]wait_for_job done: container terminated with "
                     "code: {}".format(terminated.exit_code))
        return terminated.exit_code

    def terminate_job(self, job):
        v1job = job.data
        name = v1job.metadata.name
        namespace = v1job.metadata.namespace
        grace_period_seconds = 60
        body = client.V1DeleteOptions()

        try:
            logger.debug(
                "calling delete_namespaced_pod(%s, %s, grace_period_seconds=%s)",
                name, namespace, grace_period_seconds,
            )
            self.core_api.delete_namespaced_pod(
                name, namespace, grace_period_seconds=grace_period_seconds,
                body=body,
            )
        except ApiException as e:
            logger.debug(
                "[k8s] Exception when calling " "delete_namespaced_pod: %s",
                str(e),
            )
            raise e

    def get_listen_addr(self):
        ip = None

        # if fiber.current_process() is multiprocessing.current_process():
        if not isinstance(fiber.current_process(), fiber.Process):
            # not inside docker
            # logger.debug("use interface docker0")
            ifce = "eth0"
            ip = find_ip_by_net_interface(ifce)
        else:
            # inside a Fiber process
            # logger.debug("use interface eth0")
            # ip = find_ip_by_net_interface("eth0")
            ip, ifce = find_listen_address()

        # ip = find_ip_by_net_interface(ifce)
        if ip is None:
            raise mp.ProcessError(
                "Can't find a usable IPv4 address to listen. ifce_name: {}, "
                "ifces: {}".format(ifce, psutil.net_if_addrs())
            )
        # use 0 to bind to a random free port number
        return ip, 0, ifce



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import multiprocessing as mp

import fiber.process
import fiber.config as fiber_config
import fiber.backend as fiber_backend


def init_logger(config, proc_name=None):
    logger = logging.getLogger("fiber")
    if config.log_file.lower() == "stdout":
        handler = logging.StreamHandler()
    else:
        log_dir = os.path.dirname(config.log_file)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir, exist_ok=True)

        if not proc_name:
            p = fiber.process.current_process()
            proc_name = p.name

        log_file = config.log_file + '.' + proc_name
        handler = logging.FileHandler(log_file, mode="w")

    formatter = logging.Formatter(
        '%(asctime)s %(levelname)s:%(processName)s(%(process)d):'
        '%(threadName)s(%(thread)d){%(filename)s:%(lineno)d} %(message)s')
    handler.setFormatter(formatter)
    if logger.handlers is None or len(logger.handlers) == 0:
        logger.addHandler(handler)
    else:
        logger.handlers[-1] = handler
    logger.propagate = False


def init_fiber(proc_name=None, **kwargs):
    """
    Initialize Fiber. This function is called when you want to re-initialize
    Fiber with new config values and also re-init loggers.

    :param proc_name: process name of current process
    :param kwargs: If kwargs is not None, init Fiber system with corresponding
        key/value pairs in kwargs as config keys and values.
    """
    # check backend
    if "backend" in kwargs:
        backend = kwargs["backend"]
        if (backend not in fiber_backend.available_backend
           and backend is not None):

            raise mp.ProcessError("Invalid backend: {}".format(backend))

    updates = fiber_config.init(**kwargs)

    if "log_level" in updates or "log_file" in updates:
        _config = fiber_config.get_object()
        init_logger(_config, proc_name=proc_name)



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Ring allows you to create a Ring like topology easily on a computer cluster.
"""

import multiprocessing as mp
import random

import fiber
from fiber.backend import get_backend


__all__ = [
    "Ring",
    "RingNode",
]


_manager = None


def _get_manager():
    global _manager
    if _manager is None:
        _manager = fiber.Manager()

    return _manager


# TODO(jiale) convert this into a manager type
class RingNode:
    """
    A `RingNode` represents a node in the `Ring`.

    :param rank: The id assigned to this node. Each node will be assigned a
        unique id called `rank`. Rank 0 is the control node of the `Ring`.
    """
    def __init__(self, rank):
        self.rank = rank
        self.connected = False
        self.ip = None
        self.port = None


class Ring:
    """
    Create a ring of nodes (processes). Each node runs a copy of the same
    function.

    :param processes: Number of ring nodes in this Ring.
    :param func: the target function that each ring node will run.
    :param initializer: the initialization function that each ring node runs.
        This is used to setup the custom ring structure like PyTorch's
        torch.distributed, Horovod, etc.
    :param initargs: positional arguments that are passed to initializer.
        Currently this is not used.
    """
    def __init__(self, processes, func, initializer, initargs=None):
        self.size = processes
        self.initializer = initializer
        self.initargs = initargs
        self.func = func
        self.rank = 0

        if getattr(func, "__fiber_meta__", None):
            # Propogate meta info
            # We can't set attributes to bound/unbound methods (PEP 232),
            # so we set it to Ring object
            self.__fiber_meta__ = func.__fiber_meta__

        manager = _get_manager()
        self.members = manager.list([RingNode(i) for i in range(self.size)])

    def _target(self):
        rank = self.rank
        node = self.members[rank]

        backend = get_backend()
        ip, _, _ = backend.get_listen_addr()
        port = random.randint(30000, 50000)

        node.connected = True
        node.ip = ip
        node.port = port
        self.members[rank] = node

        self.initializer(self)
        self.func(rank, self.size)

    def run(self):
        """
        Start this Ring. This will start the ring 0 process on the same machine
        and start all the other ring nodes with Fiber processes.
        """
        if self.size <= 0:
            return

        procs = []
        rank = self.rank
        # Start process rank 0
        self.rank = 0
        ctx = mp.get_context("spawn")
        p = ctx.Process(target=self._target)
        p.start()
        procs.append(p)

        for i in range(1, self.size):
            self.rank = i
            p = fiber.Process(target=self._target)
            p.start()
            procs.append(p)

        self.rank = rank
        # wait for all processes to finish
        for i in range(self.size):
            procs[i].join()



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import logging
from multiprocessing import reduction
import multiprocessing
import threading
import traceback
import select
import signal
import time


from fiber.init import init_fiber


logger = logging.getLogger('fiber')


def exit_by_signal():
    logger.info("Exiting, sending SIGTERM to current process")

    os.kill(os.getpid(), signal.SIGTERM)
    time.sleep(5)

    logger.warning("SIGTERM sent to current process but it's still running. "
                   "Trying os._exit()")
    os._exit(1)


def exit_on_fd_close(fd):
    while True:
        rl, _, _ = select.select([fd], [], [], 0)
        if fd in rl:
            logger.warning("Communication with master is closed, exiting")
            # sock is closed"
            exit_by_signal()
        time.sleep(1)


def spawn_prepare(fd):
    from_parent_r = os.fdopen(fd, "rb", closefd=False)
    preparation_data = reduction.pickle.load(from_parent_r)

    try:
        _config = preparation_data["fiber_config"]
        init_fiber(preparation_data["name"], **vars(_config))

        multiprocessing.spawn.prepare(preparation_data)
        self = reduction.pickle.load(from_parent_r)
        post_data = reduction.pickle.load(from_parent_r)

    except Exception as e:
        print("error", str(e))
        raise e

    # start watching thread so that when master quits, worker also quits
    td = threading.Thread(target=exit_on_fd_close, args=(fd,), daemon=True)
    logger.debug("Starting monitoring thread")
    td.start()

    self.ident = post_data["pid"]
    exitcode, err = self._bootstrap()
    if err:
        print("Exception in {}:".format(self))
        print(err)
        return exitcode

    return exitcode



import multiprocessing as mp
import io
import re
import time
import logging
import socket
import fiber
import fiber.popen_fiber_spawn as fspawn
import docker
import pytest


@pytest.fixture(scope="module")
def client():
    return docker.from_env()


class TestPopen():

    @pytest.fixture(autouse=True)
    def check_leak(self, request, client):
        assert fiber.active_children() == []
        if fiber.config.default_backend != "docker":
            yield
            return
        pre_workers = client.containers.list()
        yield
        post_workers = client.containers.list()
        assert pre_workers == post_workers

    def test_get_python_exe(self):
        b1 = fspawn.get_python_exe("docker")
        assert b1 == "/usr/local/bin/python"

    @pytest.mark.skip(reason="testing everytime is not needed")
    def test_command_line(self):
        fiber.init(use_bash=False)
        try:
            p = fiber.Process(name="test_command_line")
            popen = fspawn.Popen(p, backend="docker", launch=False)

            cmd = popen.get_command_line(host="127.0.0.1", port="8000", id=1)
        finally:
            fiber.init(use_bash=True)
        assert cmd == ["/usr/local/bin/python", "-c", 'import sys;import os;import socket;import struct;import fiber;import fiber.util;from multiprocessing import spawn, reduction;sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM);sock.connect(("127.0.0.1", 8000));conn = sock;conn.send(struct.pack("<I", 1));fd = conn.fileno();exitcode = fiber.util.spawn_prepare(fd);sys.exit(exitcode)', '--multiprocessing-fork']

    def test_get_job(self):
        p = fiber.Process(name="test_get_job")
        popen = fspawn.Popen(p, launch=False)

        job = popen._get_job(["echo", "hello fiber"])
        assert job.command == ["echo", "hello fiber"]
        #assert job.image == "fiber-test:latest"
        # by default, do image should be provided. Backend should decide if
        # to use default image or current container image.
        assert job.image is None

    def test_popen_method(self):
        # make sure that fiber works with multiprocessing
        p = fiber.Process(name="test_popen_method")
        popen = p._Popen(p)
        assert isinstance(popen, fiber.popen_fiber_spawn.Popen)

        p = mp.Process(name="test_popen_method2")
        popen = p._Popen(p)
        assert popen.method == "fork"
        assert isinstance(popen, mp.popen_fork.Popen)

    def test_launch_fiber_background_thread(self):
        # reset global variable
        fiber.popen_fiber_spawn._fiber_background_thread = None

        fp = io.StringIO()
        handler = logging.StreamHandler(fp)
        logger = logging.getLogger("fiber")
        logger.setLevel(level=logging.DEBUG)
        logger.addHandler(handler)

        try:
            procs = []
            for i in range(4):
                p = fiber.Process(target=time.sleep, args=(1,), name="test_launch_fiber_background_thread_{}".format(i))
                procs.append(p)
            for p in procs:
                p.start()

            logs = fp.getvalue()
            times = len(list(re.finditer('creating background thread', logs)))
            assert(times == 1)
        finally:
            for p in procs:
                p.join()
            logger.removeHandler(handler)

    def test_fiber_with_more_than_1024_fds(self):
        # select.select can't handle fd >= 1024
        # https://stackoverflow.com/questions/7328165/how-do-you-compile-python-with-1024-file-descriptors
        import resource
        resource.setrlimit(resource.RLIMIT_OFILE, (8192, 8192))

        socks = []
        for i in range(1023):
            s = socket.socket(socket.AF_INET)
            socks.append(s)

        p = fiber.Process(target=time.sleep, args=(10,), name="test_fiber_with_more_than_1024_fds")
        p.start()

        # if select is used, we should see errors like:
        # `ValueError: filedescriptor out of range in select()`
        print(p.exitcode)

        p.terminate()



#
# Module providing the `SyncManager` class for dealing
# with shared objects
#
# multiprocessing/managers.py
#
# Copyright (c) 2006-2008, R Oudkerk
# Licensed to PSF under a Contributor Agreement.
#
# Modifications Copyright (c) 2020 Uber Technologies

"""
*Managers and proxy objects* enable multiprocessing to support shared storage,
which is critical to distributed systems. Usually, this function is handled by
external storage systems like Cassandra, Redis, etc. Fiber instead provides
built-in in-memory storage for applications to use (based on
multiprocessing.managers). The interface is the same as multiprocessing's
Manager type.
"""

import logging
import multiprocessing
import multiprocessing.util as mp_util
import queue
import threading
from multiprocessing.connection import (SocketListener, _validate_family,
                                        address_type)
from multiprocessing.context import get_spawning_popen
from multiprocessing.managers import (Array, DictProxy, ListProxy, Namespace,
                                      RebuildProxy, State, Value, dispatch)
from multiprocessing.process import AuthenticationString

import fiber.util
import fiber.queues
from fiber import process
from fiber.backend import get_backend

logger = logging.getLogger('fiber')


default_family = 'AF_INET'


class Listener(multiprocessing.connection.Listener):
    def __init__(self, address=None, family=None, backlog=500, authkey=None):
        family = family or (address and address_type(address)) \
                 or default_family
        if family != 'AF_INET':
            raise NotImplementedError

        backend = get_backend()
        # TODO(jiale) Add support for other address family for
        #  backend.get_listen_addr
        address = address or backend.get_listen_addr()
        self._address = address

        _validate_family(family)
        if family == 'AF_PIPE':
            # PIPE cannot be used across machines
            raise NotImplementedError
        else:
            # Listens on '0.0.0.0' so that it accepts connection regardless of
            # net interfaces. When connect connects, it uses a specific IP
            # address.
            self._listener = SocketListener(('0.0.0.0', 0), family, backlog)

        if authkey is not None and not isinstance(authkey, bytes):
            raise TypeError('authkey should be a byte string')

        self._authkey = authkey

    # Use IP from `backend.get_listen_addr` and port from
    # self._listener._address.
    # This represents an address that other client can connect to
    address = property(lambda self: (self._address[0],
                                     self._listener._address[1]))


listener_client = {
    # Use customized Listener because we need to get listening address from
    # the backend.
    'pickle': (Listener,
               multiprocessing.connection.Client),
}


class Server(multiprocessing.managers.Server):
    def __init__(self, registry, address, authkey, serializer):
        assert isinstance(authkey, bytes)
        self.registry = registry
        self.authkey = AuthenticationString(authkey)
        Listener, Client = listener_client[serializer]

        # do authentication later
        self.listener = Listener(address=address, backlog=500)
        self.address = self.listener.address

        self.id_to_obj = {'0': (None, ())}
        self.id_to_refcount = {}
        self.id_to_local_proxy_obj = {}
        self.mutex = threading.Lock()


class BaseManager(multiprocessing.managers.BaseManager):
    """
    A Base class for people to create
    [customized managers](https://docs.python.org/3.6/library/multiprocessing.html#customized-managers).
    Equivalent to multiprocessing.managers.BaseManager but can work on a
    computer cluster.

    The API of this class is the same as
    [multiprocessing.managers.BaseManager](https://docs.python.org/3.6/library/multiprocessing.html#multiprocessing.managers.BaseManager)
    """
    _registry = {}
    _Server = Server

    def __init__(self, address=None, authkey=None, serializer='pickle',
                 ctx=None):
        if authkey is None:
            authkey = process.current_process().authkey
        self._address = address     # XXX not final address if eg ('', 0)
        self._authkey = AuthenticationString(authkey)
        self._state = State()
        self._state.value = State.INITIAL
        self._serializer = serializer
        self._Listener, self._Client = listener_client[serializer]

    def get_server(self):
        """
        Return server object with serve_forever() method and address attribute
        """
        assert self._state.value == State.INITIAL
        return Server(self._registry, self._address,
                      self._authkey, self._serializer)

    @classmethod
    def _run_server(cls, registry, address, authkey, serializer, writer,
                    initializer=None, initargs=()):
        """Create a server, report its address and run it."""
        if initializer is not None:
            initializer(*initargs)

        # create server
        server = cls._Server(registry, address, authkey, serializer)

        # inform parent process of the server's address
        writer.send(server.address)
        writer.close()

        # run the manager
        logger.info('manager serving at %r', server.address)
        server.serve_forever()

    def start(self, initializer=None, initargs=()):
        """Spawn a server process for this manager object."""
        assert self._state.value == State.INITIAL
        logger.debug("start manager %s", self)

        if initializer is not None and not callable(initializer):
            raise TypeError('initializer must be a callable')

        # pipe over which we will retrieve address of server
        reader, writer = fiber.queues.Pipe(duplex=False)

        # spawn process which runs a server
        self._process = fiber.Process(
            target=type(self)._run_server,
            args=(self._registry, self._address, self._authkey,
                  self._serializer, writer, initializer, initargs),
        )
        ident = ':'.join(str(i) for i in self._process._identity)
        self._process.name = type(self).__name__ + '-' + ident
        self._process.start()

        # get address of server
        writer.close()
        self._address = reader.recv()
        reader.close()

        # register a finalizer
        self._state.value = State.STARTED
        self.shutdown = mp_util.Finalize(
            self, type(self)._finalize_manager,
            args=(self._process, self._address, self._authkey,
                  self._state, self._Client),
            exitpriority=0
        )

    @classmethod
    def register(cls, typeid, callable=None, proxytype=None, exposed=None,
                 method_to_typeid=None, create_method=True):
        """Register a typeid with the manager type."""
        if '_registry' not in cls.__dict__:
            cls._registry = cls._registry.copy()

        if proxytype is None:
            proxytype = AutoProxy

        exposed = exposed or getattr(proxytype, '_exposed_', None)

        method_to_typeid = method_to_typeid or \
            getattr(proxytype, '_method_to_typeid_', None)

        if method_to_typeid:
            for key, value in list(method_to_typeid.items()):
                assert type(key) is str, '%r is not a string' % key
                assert type(value) is str, '%r is not a string' % value

        cls._registry[typeid] = (
            callable, exposed, method_to_typeid, proxytype
        )

        if create_method:
            def temp(self, *args, **kwds):
                logger.debug("calling self._create(%s,%s), self=%s",
                             args, kwds, self)
                token, exp = self._create(typeid, *args, **kwds)
                proxy = proxytype(
                    token, self._serializer, manager=self,
                    authkey=self._authkey, exposed=exp
                )
                conn = self._Client(token.address, authkey=self._authkey)
                dispatch(conn, None, 'decref', (token.id,))
                return proxy
            temp.__name__ = typeid
            setattr(cls, typeid, temp)


class ProcessLocalSet(set):
    def __init__(self):
        fiber.util.register_after_fork(self, lambda obj: obj.clear())

    def __reduce__(self):
        return type(self), ()


class BaseProxy(multiprocessing.managers.BaseProxy):
    """A base for proxies of shared objects."""
    _address_to_local = {}
    _mutex = fiber.util.ForkAwareThreadLock()

    def __init__(self, token, serializer, manager=None,
                 authkey=None, exposed=None, incref=True, manager_owned=False):
        with BaseProxy._mutex:
            tls_idset = BaseProxy._address_to_local.get(token.address, None)
            if tls_idset is None:
                tls_idset = fiber.util.ForkAwareLocal(), ProcessLocalSet()
                BaseProxy._address_to_local[token.address] = tls_idset

        # self._tls is used to record the connection used by this
        # thread to communicate with the manager at token.address
        self._tls = tls_idset[0]

        # self._idset is used to record the identities of all shared
        # objects for which the current process owns references and
        # which are in the manager at token.address
        self._idset = tls_idset[1]

        self._token = token
        self._id = self._token.id
        self._manager = manager
        self._serializer = serializer
        self._Client = listener_client[serializer][1]

        # Should be set to True only when a proxy object is being created
        # on the manager server; primary use case: nested proxy objects.
        # RebuildProxy detects when a proxy is being created on the manager
        # and sets this value appropriately.
        self._owned_by_manager = manager_owned

        if authkey is not None:
            self._authkey = AuthenticationString(authkey)
        elif self._manager is not None:
            self._authkey = self._manager._authkey
        else:
            self._authkey = process.current_process().authkey

        if incref:
            self._incref()

        fiber.util.register_after_fork(self, BaseProxy._after_fork)

    def connect(self):
        """Connect manager object to the server process."""
        Listener, Client = listener_client[self._serializer]
        conn = Client(self._address, authkey=self._authkey)
        dispatch(conn, None, 'dummy')
        self._state.value = State.STARTED

    def __reduce__(self):
        kwds = {}
        if get_spawning_popen() is not None:
            kwds['authkey'] = self._authkey

        if getattr(self, '_isauto', False):
            kwds['exposed'] = self._exposed_
            return (RebuildProxy,
                    (AutoProxy, self._token, self._serializer, kwds))
        else:
            return (RebuildProxy,
                    (type(self), self._token, self._serializer, kwds))


def AutoProxy(token, serializer, manager=None, authkey=None,
              exposed=None, incref=True):
    """Return an auto-proxy for `token`."""
    _Client = listener_client[serializer][1]

    if exposed is None:
        conn = _Client(token.address, authkey=authkey)
        try:
            exposed = dispatch(conn, None, 'get_methods', (token,))
        finally:
            conn.close()

    if authkey is None and manager is not None:
        authkey = manager._authkey
    if authkey is None:
        authkey = process.current_process().authkey

    ProxyType = MakeProxyType('AutoProxy[%s]' % token.typeid, exposed)
    proxy = ProxyType(token, serializer, manager=manager, authkey=authkey,
                      incref=incref)
    proxy._isauto = True
    return proxy


def MakeProxyType(name, exposed, _cache={}):
    """Return a proxy type whose methods are given by `exposed`."""
    exposed = tuple(exposed)
    try:
        return _cache[(name, exposed)]
    except KeyError:
        pass

    dic = {}

    for meth in exposed:
        exec("""def %s(self, *args, **kwds):
        return self._callmethod(%r, args, kwds)""" % (meth, meth), dic)

    ProxyType = type(name, (BaseProxy,), dic)
    ProxyType._exposed_ = exposed
    _cache[(name, exposed)] = ProxyType
    return ProxyType


# Definition of SyncManager
#

class SyncManager(BaseManager):
    """Subclass of `BaseManager` which supports a number of shared object types.

    The types that are supported include:

    * [Queue](https://docs.python.org/3.6/library/multiprocessing.html#multiprocessing.Queue)
    * [JoinableQueue](https://docs.python.org/3.6/library/multiprocessing.html#multiprocessing.JoinableQueue)
    * [list](https://docs.python.org/3.6/library/stdtypes.html#list)
    * [dict](https://docs.python.org/3.6/library/stdtypes.html#dict)
    * [Value](https://docs.python.org/3.6/library/multiprocessing.html#multiprocessing.Value)
    * [Array](https://docs.python.org/3.6/library/multiprocessing.html#multiprocessing.Array)
    * [Namespace](https://docs.python.org/3.6/library/multiprocessing.html#multiprocessing.managers.Namespace)
    * [AsyncResult](https://docs.python.org/3.6/library/multiprocessing.html#multiprocessing.pool.AsyncResult)

    `SyncManager` is exposed as `fiber.Manager`. You can also use
    `fiber.managers.SyncManager` directly. The difference is that
    `fiber.Manager` will start the newly created manager and when using
    `SyncManager`, the user need to call the `start()` method of the manager
    explicitly.

    `fiber.Manager` example:

    ```
    manager = fiber.Manager()
    d = manager.dict()
    d["name"] = "managed_dict"
    print(d)
    ```

    `fiber.managers.SyncManager` example:

    ```
    manager = fiber.managers.SyncManager()
    manager.start()
    l = manager.list()
    l.append("x")
    print(l)
    ```
    """


ArrayProxy = MakeProxyType('ArrayProxy', (
    '__len__', '__getitem__', '__setitem__'
))


class NamespaceProxy(BaseProxy):
    _exposed_ = ('__getattribute__', '__setattr__', '__delattr__')

    def __getattr__(self, key):
        if key[0] == '_':
            return object.__getattribute__(self, key)
        callmethod = object.__getattribute__(self, '_callmethod')
        return callmethod('__getattribute__', (key,))

    def __setattr__(self, key, value):
        if key[0] == '_':
            return object.__setattr__(self, key, value)
        callmethod = object.__getattribute__(self, '_callmethod')
        return callmethod('__setattr__', (key, value))

    def __delattr__(self, key):
        if key[0] == '_':
            return object.__delattr__(self, key)
        callmethod = object.__getattribute__(self, '_callmethod')
        return callmethod('__delattr__', (key,))


class ValueProxy(BaseProxy):
    _exposed_ = ('get', 'set')

    def get(self):
        return self._callmethod('get')

    def set(self, value):
        return self._callmethod('set', (value,))

    value = property(get, set)


# Define AsyncProxy and AsyncManager

class AsyncProxyResult():
    def __init__(self, conn, proxy):
        self._conn = conn
        self._proxy = proxy

    def get(self):
        kind, result = self._conn.recv()

        if kind == '#RETURN':
            return result
        elif kind == '#PROXY':
            # TODO(jiale)
            raise NotImplementedError


class AsyncBaseProxy(BaseProxy):
    def _callmethod(self, methodname, args=(), kwds={}):
        try:
            conn = self._tls.connection
        except AttributeError:
            self._connect()
            conn = self._tls.connection

        conn.send((self._id, methodname, args, kwds))

        return AsyncProxyResult(conn, self)


def MakeAsyncProxyType(name, exposed, _cache={}):
    """Return a proxy type whose methods are given by `exposed`."""
    exposed = tuple(exposed)
    try:
        return _cache[(name, exposed)]
    except KeyError:
        pass

    dic = {}

    for meth in exposed:
        exec("""def %s(self, *args, **kwds):
        return self._callmethod(%r, args, kwds)""" % (meth, meth), dic)

    ProxyType = type(name, (AsyncBaseProxy,), dic)
    ProxyType._exposed_ = exposed
    _cache[(name, exposed)] = ProxyType
    return ProxyType


def AsyncAutoProxy(token, serializer, manager=None, authkey=None,
                   exposed=None, incref=True):
    """Return an auto-proxy for `token`."""
    _Client = listener_client[serializer][1]

    if exposed is None:
        conn = _Client(token.address, authkey=authkey)
        try:
            exposed = dispatch(conn, None, 'get_methods', (token,))
        finally:
            conn.close()

    if authkey is None and manager is not None:
        authkey = manager._authkey
    if authkey is None:
        authkey = process.current_process().authkey

    ProxyType = MakeAsyncProxyType('AutoProxy[%s]' % token.typeid, exposed)
    proxy = ProxyType(token, serializer, manager=manager, authkey=authkey,
                      incref=incref)
    proxy._isauto = True
    return proxy


class AsyncValueProxy(AsyncBaseProxy):
    _exposed_ = ('get', 'set')

    def get(self):
        return self._callmethod('get')

    def set(self, value):
        return self._callmethod('set', (value,))

    value = property(get, set)


class AsyncManager(BaseManager):
    """
    Asynchronous version of BaseManager. Currently, this type doesn't register
    any shared object types like SyncManager, it is more meant to be used
    as a base class for
    [customized managers](https://docs.python.org/3.6/library/multiprocessing.html#customized-managers).

    The API of the class is the same as `SyncManager` and
    [multiprocessing.managers.BaseManager](https://docs.python.org/3.6/library/multiprocessing.html#multiprocessing.managers.BaseManager)

    Example usage:

    ```python
    from fiber.managers import AsyncManager

    class Calculator():
        def add(self, a, b):
            return a + b

    class MyManager(AsyncManager):
        pass

    MyManager.register("Calculator", Calculator)
    manager = MyManager()
    manager.start()

    calc = manager.Calculator()
    res = calc.add(10, 32)
    print(type(res)) # fiber.managers.AsyncProxyResult
    print(res.get()) # 42
    ```
    """
    @classmethod
    def register(cls, typeid, callable=None, proxytype=None, exposed=None,
                 method_to_typeid=None, create_method=True):
        """Register a typeid with the manager type."""
        if '_registry' not in cls.__dict__:
            cls._registry = cls._registry.copy()

        if proxytype is None:
            proxytype = AsyncAutoProxy

        exposed = exposed or getattr(proxytype, '_exposed_', None)

        method_to_typeid = method_to_typeid or \
            getattr(proxytype, '_method_to_typeid_', None)

        if method_to_typeid:
            for key, value in list(method_to_typeid.items()):
                assert type(key) is str, '%r is not a string' % key
                assert type(value) is str, '%r is not a string' % value

        cls._registry[typeid] = (
            callable, exposed, method_to_typeid, proxytype
        )

        if create_method:
            def temp(self, *args, **kwds):
                logger.debug("calling self._create(%s,%s)", args, kwds)
                token, exp = self._create(typeid, *args, **kwds)
                proxy = proxytype(
                    token, self._serializer, manager=self,
                    authkey=self._authkey, exposed=exp
                )
                conn = self._Client(token.address, authkey=self._authkey)
                dispatch(conn, None, 'decref', (token.id,))
                return proxy
            temp.__name__ = typeid
            setattr(cls, typeid, temp)


AsyncBaseListProxy = MakeAsyncProxyType('AsyncBaseListProxy', (
    '__add__', '__contains__', '__delitem__', '__getitem__', '__len__',
    '__mul__', '__reversed__', '__rmul__', '__setitem__',
    'append', 'count', 'extend', 'index', 'insert', 'pop', 'remove',
    'reverse', 'sort', '__imul__'
    ))


class AsyncListProxy(AsyncBaseListProxy):
    def __iadd__(self, value):
        self._callmethod('extend', (value,))
        return self

    def __imul__(self, value):
        self._callmethod('__imul__', (value,))
        return self


AsyncDictProxy = MakeAsyncProxyType('AsyncDictProxy', (
    '__contains__', '__delitem__', '__getitem__', '__iter__', '__len__',
    '__setitem__', 'clear', 'copy', 'get', 'has_key', 'items',
    'keys', 'pop', 'popitem', 'setdefault', 'update', 'values'
    ))
AsyncDictProxy._method_to_typeid_ = {
    '__iter__': 'Iterator',
    }


AsyncArrayProxy = MakeAsyncProxyType('AsyncArrayProxy', (
    '__len__', '__getitem__', '__setitem__'
    ))

# Register methods for SyncManager

SyncManager.register('Queue', queue.Queue)
SyncManager.register('JoinableQueue', queue.Queue)
# TODO(jiale) add support for these types
# SyncManager.register('Event', threading.Event, EventProxy)
# SyncManager.register('Lock', threading.Lock, AcquirerProxy)
# SyncManager.register('RLock', threading.RLock, AcquirerProxy)
# SyncManager.register('Semaphore', threading.Semaphore, AcquirerProxy)
# SyncManager.register('BoundedSemaphore', threading.BoundedSemaphore,
#                     AcquirerProxy)
# SyncManager.register('Condition', threading.Condition, ConditionProxy)
# SyncManager.register('Barrier', threading.Barrier, BarrierProxy)
# SyncManager.register('Pool', pool.Pool, PoolProxy)
SyncManager.register('list', list, ListProxy)
SyncManager.register('dict', dict, DictProxy)
SyncManager.register('Value', Value, ValueProxy)
SyncManager.register('Array', Array, ArrayProxy)
SyncManager.register('Namespace', Namespace, NamespaceProxy)

# types returned by methods of PoolProxy
# SyncManager.register('Iterator', proxytype=IteratorProxy, create_method=False) # noqa E501
SyncManager.register('AsyncResult', create_method=False)


# Register methods for AyncManager

#AsyncManager.register('Value', Value, AsyncValueProxy)
#AsyncManager.register('Queue', queue.Queue)
#AsyncManager.register('JoinableQueue', queue.Queue)
#AsyncManager.register('list', list, AsyncListProxy)
#AsyncManager.register('dict', dict, AsyncDictProxy)
#AsyncManager.register('Value', Value, ValueProxy)
#AsyncManager.register('Array', Array, AsyncArrayProxy)
#AsyncManager.register('Namespace', Namespace, NamespaceProxy)



import fiber
import fiber.config
import logging
import pytest
import time

import fiber.config as fiber_config


def config_worker(writer):
    cfg = fiber.config.get_dict()
    print("[worker]", cfg)
    writer.send(cfg)
    time.sleep(0.1)


class TestMisc():
    def test_config_file(self):
        config = fiber.config.Config(conf_file="tests/demo_config")
        assert config.image == "demo-image:latest"
        assert config.log_level == logging.INFO
        assert config.ipc_active is False
        assert config.ipc_admin_master_port == 0
        assert config.ipc_admin_worker_port == 8000
        assert config.mem_per_job == 12345

    def test_invalid_key(self):
        with pytest.raises(ValueError, match=r""):
            _ = fiber.config.Config(conf_file="tests/invalid_config")

    def test_config_env(self):
        import os
        old_val = os.environ.get("FIBER_IMAGE", None)
        os.environ["FIBER_IMAGE"] = "image-from-env:latest"
        try:
            fiber.init()
            assert fiber_config.image == "image-from-env:latest"
        finally:
            if old_val is None:
                del os.environ["FIBER_IMAGE"]
            else:
                os.environ["FIBER_IMAGE"] = old_val
            fiber.init()

    def test_config_sync(self):
        # child process should have the same config as parent process
        r, w = fiber.Pipe()
        p = fiber.Process(target=config_worker, args=(w,))
        p.start()

        got = r.recv()
        expected = fiber.config.get_dict()

        p.join()
        assert got == expected



#
# Module providing the `Pool` class for managing a process pool
#
# multiprocessing/pool.py
#
# Copyright (c) 2006-2008, R Oudkerk
# Licensed to PSF under a Contributor Agreement.
#
# Modifications Copyright (c) 2020 Uber Technologies

"""
*Pools* are supported by Fiber. They allow the user to manage a pool of
worker processes. Fiber extend pools with *job-backed processes* so that it can
manage thousands of (remote) workers per pool. Users can also create multiple
pools at the same time.

Fiber implements 2 different version of `Pool`: `ZPool` and `ResilientZPool`.
Both has the same API as `multiprocessing.Pool`. `ZPool` is pool based on
"r"/"w" socket pairs.
`ResilientZPool` is `ZPool` + [error handling](advanced.md#error-handling).
Failed tasks will be resubmitted to the Pool and worked on by other pool
workers.

By default, `ResilientZPool` is exposed as `fiber.Pool`.

Example:

```python
pool = fiber.Pool(processes=4)
pool.map(math.sqrt, range(10))
```
"""

import logging
import multiprocessing as mp
import multiprocessing.pool as mp_pool
import multiprocessing.util as mp_util
import math
import queue
import random
import struct
import threading
import sys
import time
import secrets
import traceback
from multiprocessing.pool import (CLOSE, RUN, TERMINATE,
                                  ExceptionWithTraceback, MaybeEncodingError,
                                  ThreadPool, _helper_reraises_exception)

import fiber.queues
import fiber.config as config
from fiber.backend import get_backend
from fiber.queues import LazyZConnection
from fiber.socket import Socket
from fiber.process import current_process
import signal


if fiber.util.is_in_interactive_console():
    import cloudpickle as pickle
else:
    import pickle

logger = logging.getLogger('fiber')

MIN_PORT = 40000
MAX_PORT = 65535


def safe_join_worker(proc):
    p = proc
    if p.is_alive():
        # worker has not yet exited
        logger.debug('cleaning up worker %s' % p.pid)

        p.join(5)


def safe_terminate_worker(proc):
    delay = random.random()

    # Randomize start time to prevent overloading the server
    logger.debug(
        "start multiprocessing.pool.worker terminator thread for proc %s "
        "with delay %s", proc.name, delay)
    time.sleep(delay)

    p = proc
    if p.exitcode is None:
        p.terminate()

    logger.debug("safe_terminate_worker() finished")


def safe_start(proc):
    try:
        proc.start()
        proc._start_failed = False
    except Exception:
        msg = traceback.format_exc()
        logging.warning("failed to start process %s: %s", proc.name, msg)
        # Set this so that this process can be cleaned up later
        proc._start_failed = True


def mp_worker_core(inqueue, outqueue, maxtasks=None, wrap_exception=False):
    logger.debug('mp_worker_core running')
    put = outqueue.put
    get = inqueue.get

    completed = 0
    while maxtasks is None or (maxtasks and completed < maxtasks):
        try:
            task = get()
        except (EOFError, OSError):
            logger.debug('worker got EOFError or OSError -- exiting')
            break

        if task is None:
            logger.debug('worker got sentinel -- exiting')
            break

        job, i, func, args, kwds = task
        try:
            result = (True, func(*args, **kwds))
        except Exception as e:
            if wrap_exception and func is not _helper_reraises_exception:
                e = ExceptionWithTraceback(e, e.__traceback__)
            result = (False, e)
        try:
            put((job, i, result))
        except Exception as e:
            wrapped = MaybeEncodingError(e, result[1])
            logger.debug("Possible encoding error while sending result: %s" % (
                wrapped))
            put((job, i, (False, wrapped)))

        task = job = result = func = args = kwds = None
        completed += 1
    logger.debug('worker exiting after %s tasks' % completed)


def mp_worker(inqueue, outqueue, initializer=None, initargs=(), maxtasks=None,
              wrap_exception=False, num_workers=1):
    """This is mostly the same as multiprocessing.pool.worker, the difference
    is that it will start multiple workers (specified by `num_workers` argument)
    via multiproccessing and allow the Fiber pool worker to take multiple CPU
    cores.
    """

    assert maxtasks is None or (type(maxtasks) == int and maxtasks > 0)
    if hasattr(inqueue, '_writer'):
        inqueue._writer.close()
        outqueue._reader.close()

    if initializer is not None:
        initializer(*initargs)

    workers = []
    for i in range(num_workers):
        # Use fork context to make sure that imported modules and other things
        # are shared. This is useful when modules share objects between
        # processes.
        ctx = mp.get_context('fork')
        p = ctx.Process(target=mp_worker_core,
                        args=(inqueue, outqueue, maxtasks, wrap_exception))
        p.start()
        workers.append(p)

    for w in workers:
        w.join()


class ClassicPool(mp_pool.Pool):

    @staticmethod
    def Process(ctx, *args, **kwds):
        return fiber.process.Process(*args, **kwds)

    def __init__(self, processes=None, initializer=None, initargs=(),
                 maxtasksperchild=None, cluster=None):

        self._ctx = None
        self._setup_queues()
        self._taskqueue = queue.Queue()
        self._cache = {}
        self._state = RUN
        self._maxtasksperchild = maxtasksperchild
        self._initializer = initializer
        self._initargs = initargs

        if processes is None:
            processes = 1
        if processes < 1:
            raise ValueError("Number of processes must be at least 1")

        if initializer is not None and not callable(initializer):
            raise TypeError("initializer must be a callable")

        self._processes = processes
        self._pool = []
        self._threads = []
        self._repopulate_pool()

        # Worker handler
        self._worker_handler = threading.Thread(
            target=ClassicPool._handle_workers,
            args=(self._cache, self._taskqueue, self._ctx, self.Process,
                  self._processes, self._pool, self._threads, self._inqueue,
                  self._outqueue, self._initializer, self._initargs,
                  self._maxtasksperchild, self._wrap_exception)
        )
        self._worker_handler.daemon = True
        self._worker_handler._state = RUN
        self._worker_handler.start()
        logger.debug(
            "Pool: started _handle_workers thread(%s:%s)",
            self._worker_handler.name, self._worker_handler.ident
        )

        # Task handler

        # NOTE: Fiber socket is not thread safe. SimpleQueue is built with Fiber
        # socket, so it shouldn't be shared between different threads. But
        # here _quick_put and _inqueue are only accessed by _task_handler
        # except when the Pool is terminated and `_terminate` is called.
        self._task_handler = threading.Thread(
            target=ClassicPool._handle_tasks,
            args=(self._taskqueue, self._quick_put, self._outqueue,
                  self._pool, self._cache)
        )
        self._task_handler.daemon = True
        self._task_handler._state = RUN
        self._task_handler.start()
        logger.debug(
            "Pool: started _handle_tasks thread(%s:%s)",
            self._task_handler.name, self._task_handler.ident
        )

        # Result handler
        self._result_handler = threading.Thread(
            target=ClassicPool._handle_results,
            args=(self._outqueue, self._quick_get, self._cache)
        )
        self._result_handler.daemon = True
        self._result_handler._state = RUN
        self._result_handler.start()
        logger.debug(
            "Pool: started _handle_results thread(%s:%s)",
            self._result_handler.name, self._result_handler.ident
        )

        # TODO use fiber's own weak ref
        self._terminate = mp_util.Finalize(
            self, self._terminate_pool,
            args=(self._taskqueue, self._inqueue, self._outqueue, self._pool,
                  self._threads, self._worker_handler, self._task_handler,
                  self._result_handler, self._cache),
            exitpriority=15
        )
        logger.debug("Pool: registered _terminate_pool finalizer")

    def _setup_queues(self):
        self._inqueue = fiber.queues.SimpleQueue()
        logger.debug("Pool|created Pool._inqueue: %s", self._inqueue)
        self._outqueue = fiber.queues.SimpleQueue()
        logger.debug("Pool|created Pool._outqueue: %s", self._outqueue)

        # TODO(jiale) use send_string instead?
        self._quick_put = self._inqueue.put
        # TODO(jiale) can't use _outqueue.reader.get because _outqueue.reader
        # is a REQ socket. It can't be called consecutively.
        self._quick_get = self._outqueue.get

    def _map_async(self, func, iterable, mapper, chunksize=None, callback=None,
                   error_callback=None):
        """
        Helper function to implement map, starmap and their async counterparts.
        """
        if self._state != RUN:
            raise ValueError("Pool not running")
        if not hasattr(iterable, '__len__'):
            iterable = list(iterable)

        if chunksize is None:
            # use self._processes to replace len(self._pool) to calculate
            # chunk size. This is because len(self._pool) is the number
            # of jobs not total processes.
            chunksize, extra = divmod(len(iterable), self._processes * 4)
            if extra:
                chunksize += 1
        if len(iterable) == 0:
            chunksize = 0

        task_batches = ClassicPool._get_tasks(func, iterable, chunksize)
        result = mp_pool.MapResult(self._cache, chunksize, len(iterable),
                                   callback, error_callback=error_callback)
        self._taskqueue.put(
            (
                self._guarded_task_generation(result._job,
                                              mapper,
                                              task_batches),
                None
            )
        )
        return result

    @staticmethod
    def _handle_workers(cache, taskqueue, ctx, Process, processes, pool,
                        threads, inqueue, outqueue, initializer, initargs,
                        maxtasksperchild, wrap_exception):
        thread = threading.current_thread()

        # Keep maintaining workers until the cache gets drained, unless the
        # pool is terminated.
        while thread._state == RUN or (cache and thread._state != TERMINATE):
            ClassicPool._maintain_pool(ctx, Process, processes, pool, threads,
                                inqueue, outqueue, initializer, initargs,
                                maxtasksperchild, wrap_exception)
            time.sleep(0.1)

        logger.debug("_handle_workers exits")

    @staticmethod
    def _join_exited_workers(pool):
        """Cleanup after any worker processes which have exited due to reaching
        their specified lifetime.  Returns True if any workers were cleaned up.
        """
        cleaned = False

        thread = threading.current_thread()
        for i in reversed(range(len(pool))):
            # leave dead workers for later cleaning
            if getattr(thread, "_state", None) == TERMINATE:
                logger.debug("pool is being terminated, "
                             "leave dead workers for later cleaning")
                break

            worker = pool[i]
            if worker._start_failed:
                cleaned = True
                del pool[i]
                logger.debug("remove process %s which failed to "
                             "start", worker.name)
                continue

            logger.debug("check worker.exitcode %s", worker.name)
            if worker.exitcode is not None:
                # worker exited
                worker.join()
                cleaned = True
                del pool[i]
        return cleaned

    def _repopulate_pool(self):
        return self._repopulate_pool_static(self._ctx, self.Process,
                                            self._processes,
                                            self._pool, self._threads,
                                            self._inqueue,
                                            self._outqueue, self._initializer,
                                            self._initargs,
                                            self._maxtasksperchild,
                                            self._wrap_exception)

    @staticmethod
    def _repopulate_pool_static(ctx, Process, processes, pool, threads,
                                inqueue, outqueue, initializer, initargs,
                                maxtasksperchild, wrap_exception):
        """Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        logger.debug("_repolulate_pool_static, pool: %s", pool)

        thread = threading.current_thread()

        workers_per_fp = config.cpu_per_job

        remain = processes - len(pool) * workers_per_fp
        while remain > 0:
            # don't repolulate workers if exiting
            logger.debug(
                "processes %s, len(pool) %s",
                processes, len(pool)
            )
            if getattr(thread, "_state", None) == TERMINATE:
                logger.debug("pool is being terminated, stop "
                             "repopulating workers")
                break

            num_workers = (
                workers_per_fp if remain >= workers_per_fp else remain
            )

            w = Process(ctx, target=mp_worker,
                        args=(inqueue, outqueue,
                              initializer,
                              initargs, maxtasksperchild,
                              wrap_exception, num_workers))
            pool.append(w)
            remain = remain - num_workers
            w.name = w.name.replace('Process', 'PoolWorker')
            w.daemon = True
            td = threading.Thread(target=safe_start, args=(w,))
            td.start()
            threads.append(td)
            logger.debug("start multiprocessing.pool.worker starter "
                         "thread(%s:%s)",
                         td.name, td.ident)

        logger.debug("_repolulate_pool_static done, pool: %s", pool)

    @staticmethod
    def _maintain_pool(ctx, Process, processes, pool, threads, inqueue,
                       outqueue, initializer, initargs, maxtasksperchild,
                       wrap_exception):
        """Clean up any exited workers and start replacements for them.
        """
        if ClassicPool._join_exited_workers(pool):
            ClassicPool._repopulate_pool_static(ctx, Process, processes, pool,
                                         threads, inqueue, outqueue,
                                         initializer, initargs,
                                         maxtasksperchild,
                                         wrap_exception)

    @staticmethod
    def _handle_tasks(taskqueue, put, outqueue, pool, cache):
        thread = threading.current_thread()

        for taskseq, set_length in iter(taskqueue.get, None):
            task = None
            try:
                # iterating taskseq cannot fail
                for task in taskseq:
                    if thread._state:
                        break
                    try:
                        put(task)
                    except Exception as e:
                        job, idx = task[:2]
                        try:
                            cache[job]._set(idx, (False, e))
                        except KeyError:
                            pass
                else:
                    if set_length:
                        idx = task[1] if task else -1
                        set_length(idx + 1)
                    continue
                break
            finally:
                task = taskseq = job = None
        else:
            logger.debug('_handle_tasks: task handler got sentinel')

        try:
            # tell result handler to finish when cache is empty
            logger.debug('_handle_tasks: task handler sending sentinel '
                         'to result handler')
            outqueue.put(None)

            # tell workers there is no more work
            logger.debug('_handle_tasks: task handler sending sentinel '
                         'to workers')
            for p in pool:
                put(None)
        except OSError:
            logger.debug('_handle_tasks: task handler got OSError when '
                         'sending sentinels')

        logger.debug('_handle_tasks: task handler exiting')

    @staticmethod
    def _handle_results(outqueue, get, cache):
        thread = threading.current_thread()

        while 1:
            try:
                task = get()
            except (OSError, EOFError):
                # logger.debug('result handler got EOFError/OSError: exiting')
                return

            if thread._state:
                assert thread._state == TERMINATE
                # logger.debug('result handler found thread._state=TERMINATE')
                break

            if task is None:
                # logger.debug('result handler got sentinel')
                break

            job, i, obj = task
            try:
                cache[job]._set(i, obj)
            except KeyError:
                pass
            task = job = obj = None

        while cache and thread._state != TERMINATE:
            try:
                task = get()
            except (OSError, EOFError):
                logger.debug('result handler got EOFError/OSError -- exiting')
                return

            if task is None:
                logger.debug('result handler ignoring extra sentinel')
                continue
            job, i, obj = task
            try:
                cache[job]._set(i, obj)
            except KeyError:
                pass
            task = job = obj = None

        if hasattr(outqueue, '_reader'):
            logger.debug('ensuring that outqueue is not full')
            # If we don't make room available in outqueue then
            # attempts to add the sentinel (None) to outqueue may
            # block.  There is guaranteed to be no more than 2 sentinels.
            try:
                for i in range(10):
                    if not outqueue._reader.poll():
                        break
                    get()
            except (OSError, EOFError):
                pass

        logger.debug('result handler exiting: len(cache)=%s, thread._state=%s',
                     len(cache), thread._state)

    @classmethod
    def _terminate_pool(cls, taskqueue, inqueue, outqueue, pool, threads,
                        worker_handler, task_handler, result_handler, cache):
        # this is guaranteed to only be called once
        logger.debug('finalizing pool')
        start = time.time()
        terminate_start = start

        worker_handler._state = TERMINATE
        task_handler._state = TERMINATE

        logger.debug('helping task handler/workers to finish')

        assert result_handler.is_alive() or len(cache) == 0

        result_handler._state = TERMINATE
        outqueue.put(None)                  # sentinel

        # This was previously done in _handle_workers, we move it here to
        # improve terminating speed
        logger.debug("send sentinel, put None in task queue")
        taskqueue.put(None)

        elapsed = time.time() - start
        start = start + elapsed
        logger.debug("outqueue put None took %s", elapsed)

        # Since worker_handler._state is already set, not more new processes
        # will be created. So we don't need to wait for worker_handler.
        # Original comments below doesn't apply anymore. They are kept here
        # for reference.
        #
        # Original comment: We must wait for the worker handler to exit before
        # terminating workers because we don't want workers to be restarted
        # behind our back.

        # logger.debug('joining worker handler')
        # if threading.current_thread() is not worker_handler:
        #     worker_handler.join()

        elapsed = time.time() - start
        start = start + elapsed
        logger.debug("joining worker handler took %s", elapsed)

        logger.debug('tell process._popen that it should exit')
        for p in pool:
            if p._popen is not None:
                logger.debug('set process._popen._exiting = True '
                             'for %s', p.name)
                p._popen._exiting = True

        elapsed = time.time() - start
        start = start + elapsed
        logger.debug("setting p._popen._exiting took %s", elapsed)

        logger.debug('joining starter threads')
        for td in threads:
            # wait for all starter threads to finish
            logger.debug('joining starter thread %s', td.name)
            td.join()
            logger.debug('joining starter thread finished')

        elapsed = time.time() - start
        start = start + elapsed
        logger.debug("joining starter threads took %s", elapsed)

        N = min(100, len(pool))
        # Thread pool used for process termination. With randomized delay, this
        # roughly equals N requests per second.
        tpool = ThreadPool(N)

        # Terminate workers which haven't already finished.
        if pool and hasattr(pool[0], 'terminate'):
            logger.debug('terminating workers')
            tpool.map_async(safe_terminate_worker, pool)

        elapsed = time.time() - start
        start = start + elapsed
        logger.debug("run safe_terminate_worker threads took %s", elapsed)

        logger.debug('joining task handler')
        if threading.current_thread() is not task_handler:
            task_handler.join()

        elapsed = time.time() - start
        start = start + elapsed
        logger.debug("joining task handler took %s", elapsed)

        logger.debug('joining result handler')
        if threading.current_thread() is not result_handler:
            result_handler.join()

        elapsed = time.time() - start
        start = start + elapsed
        logger.debug("joining result handler took %s", elapsed)

        tpool = ThreadPool(N)
        if pool and hasattr(pool[0], 'terminate'):
            logger.debug('joining pool workers, this may take some '
                         'time. workers: %s', pool)
            tpool.map(safe_join_worker, pool)

        elapsed = time.time() - start
        start = start + elapsed
        logger.debug("joining pool workers took %s", elapsed)
        logger.debug(
            "terminating pool finished. it took %s",
            time.time() - terminate_start
        )


class Inventory():
    """An inventory object that holds map results.

    An `seq` needs to be requested with `add` method so that map result can
    be tracked. `seq` acts as an request id. In later stage, `get` method
    can be called with corresponding `seq`. This inventory will handle waiting,
    managing results from different map calls.
    """
    def __init__(self, queue_get):
        self._seq = 0
        self._queue_get = queue_get
        self._inventory = {}
        self._spec = {}
        self._idx_cur = {}

    def add(self, ntasks):
        self._seq += 1
        self._inventory[self._seq] = [None] * ntasks
        self._spec[self._seq] = ntasks
        self._idx_cur[self._seq] = 0
        return self._seq

    def get(self, job_seq):
        n = self._spec[job_seq]

        while n != 0:
            # seq, batch, batch + i, result
            seq, _, i, result = self._queue_get()
            self._inventory[seq][i] = result
            self._spec[seq] -= 1
            if seq == job_seq:
                n = self._spec[seq]

        ret = self._inventory[job_seq]
        self._inventory[job_seq] = None
        return ret

    def iget_unordered(self, job_seq):
        n = self._spec[job_seq]

        while n != 0:
            # seq, batch, batch + i, result
            seq, _, i, result = self._queue_get()
            self._inventory[seq][i] = result
            self._spec[seq] -= 1
            if seq == job_seq:
                res = self._inventory[job_seq][i]
                self._inventory[job_seq][i] = None
                n = self._spec[seq]
                yield res

        return

    def iget_ordered(self, job_seq):
        idx = self._idx_cur[job_seq]
        total = len(self._inventory[job_seq])

        while idx != total:
            if self._inventory[job_seq][idx] is not None:

                res = self._inventory[job_seq][idx]
                self._inventory[job_seq][idx] = None

                idx += 1
                self._spec[job_seq] = idx
                yield res

                continue

            # seq, batch, batch + i, result
            seq, _, i, result = self._queue_get()
            self._inventory[seq][i] = result

            if seq == job_seq:
                if i == idx:
                    # got the next one
                    res = self._inventory[job_seq][i]
                    self._inventory[job_seq][i] = None

                    idx += 1
                    self._spec[job_seq] = idx

                    yield res

        return


class MapResult():
    def __init__(self, seq, inventory):
        self._seq = seq
        self._inventory = inventory

    def get(self):
        return self._inventory.get(self._seq)

    def iget_ordered(self):
        return self._inventory.iget_ordered(self._seq)

    def iget_unordered(self):
        return self._inventory.iget_unordered(self._seq)


class ApplyResult(MapResult):
    """An object that is returned by asynchronous methods of `Pool`. It
    represents an handle that can be used to get the actual result.
    """

    def get(self):
        """Get the actual result represented by this object

        :returns: Actual result. This method will block if the actual result
            is not ready.
        """
        return self._inventory.get(self._seq)[0]


def zpool_worker_core(master_conn, result_conn, maxtasksperchild,
                      wrap_exception, rank=-1, req=False):
    """
    The actual function that processes tasks.

    :param master_conn: connection that is used to read task from Pool.
    :param result_conn: connection that is used to send results to Pool.
    :param maxtasksperchild: TODO: max tasks per child process.
    "param wrap_exception: TODO
    "param rank: used when zpool_worker started many zpool_worker_core
        processes. `rank` is the id of this process.
    "pram req": whether master_conn is based on a REQ type socket. When this
        flag is set, `zpool_worker_core` needs to request tasks from the Pool
        by send it's (ident, proc.id) to the Pool and Pool will send a task
        back to the worker. This doesn't change result_conn.
    """
    logger.debug("zpool_worker_core started %s", rank)

    proc = None
    ident = secrets.token_bytes(4)
    if req:
        proc = current_process()

    while True:
        if req:
            # master_conn is a REQ type socket, need to send id (rank) to
            # master to request a task. id is packed in type unsigned short (H)
            master_conn.send_bytes(struct.pack("4si", ident, proc.pid))
            #print("worker_core send", ident, proc.pid)

        task = master_conn.recv()
        if task is None:
            logger.debug("task worker got None, exiting")
            break

        seq, batch, func, arg_list, starmap = task
        logger.debug('zpool_worker got %s, %s, %s, %s', seq, batch,
                     func, arg_list)
        if len(arg_list) == 0:
            continue

        if starmap:
            for i, arg_item in enumerate(arg_list):
                nargs = len(arg_item)
                if nargs == 2:
                    args, kwds = arg_item
                    res = func(*args, **kwds)
                elif nargs == 1:
                    args = arg_item[0]
                    res = func(*args)
                else:
                    raise ValueError("Bad number of args, %s %s",
                                     nargs, arg_item)

                data = (seq, batch, batch + i, res)
                if req:
                    data += (ident,)
                result_conn.send(data)
        else:
            for i, args in enumerate(arg_list):
                res = func(args)
                data = (seq, batch, batch + i, res)
                if req:
                    data += (ident,)
                result_conn.send(data)
    #print("worker_core exit, ", rank, proc.pid)


def handle_signal(signal, frame):
    # run sys.exit() so that atexit handlers can run
    sys.exit()

def zpool_worker(master_conn, result_conn, initializer=None, initargs=(),
                 maxtasks=None, wrap_exception=False, num_workers=1, req=False):
    """
    The entry point of Pool worker function.

    :param master_conn: connection that is used to read task from Pool.
    :param result_conn: connection that is used to send results to Pool.
    :param maxtasksperchild: TODO: max tasks per child process.
    :param wrap_exception: TODO
    :param num_workers: number of workers to start.
    :param rank: used when zpool_worker started many zpool_worker_core
        processes. `rank` is the id of this process.
    :param req: whether master_conn is based on a REQ type socket. When this
        flag is set, `zpool_worker_core` needs to request tasks from the Pool
        by send it's (ident, proc.id) to the Pool and Pool will send a task
        back to the worker. This doesn't change result_conn.
    """
    logger.debug("zpool_worker running")
    signal.signal(signal.SIGTERM, handle_signal)

    if wrap_exception:
        # TODO(jiale) implement wrap_exception
        raise NotImplementedError

    assert maxtasks is None or (type(maxtasks) == int and maxtasks > 0)

    if initializer is not None:
        initializer(*initargs)

    if num_workers == 1:
        return zpool_worker_core(master_conn, result_conn, maxtasks,
                                 wrap_exception, 0, req=req)

    workers = []
    for i in range(num_workers):
        # Use fork context to make sure that imported modules and other things
        # are shared. This is useful when modules share objects between
        # processes.
        ctx = mp.get_context('fork')
        p = ctx.Process(target=zpool_worker_core,
                        args=(master_conn, result_conn, maxtasks,
                              wrap_exception, i, req), daemon=True)
        p.start()
        workers.append(p)

    for w in workers:
        w.join()


class ZPool():
    """A Pool implementation based on Fiber sockets.

    ZPool directly uses Fiber sockets instead of SimpleQueue for tasks and
    results handling. This makes it faster.
    """

    def __init__(self, processes=None, initializer=None, initargs=(),
                 maxtasksperchild=None, cluster=None,
                 master_sock_type="w"):

        self._pool = []
        # Set default processes to 1
        self._processes = processes if processes is not None else 1
        self._initializer = initializer
        self._initargs = initargs
        self._maxtasksperchild = maxtasksperchild
        self._cluster = cluster
        self._seq = 0
        self._state = RUN
        self.taskq = queue.Queue()
        self.sent_tasks = 0
        self.recv_tasks = 0
        self.max_processing_tasks = 20000

        # networking related
        backend = get_backend()
        ip, _, _ = backend.get_listen_addr()

        socket = Socket(mode=master_sock_type)
        _master_port = socket.bind()
        _master_addr = 'tcp://{}:{}'.format(ip, _master_port)
        self._master_addr = _master_addr
        self._master_sock = socket

        socket = Socket(mode="r")
        _result_port = socket.bind()
        _result_addr = 'tcp://{}:{}'.format(ip, _result_port)
        self._result_addr = _result_addr
        self._result_sock = socket

        logger.debug("creating %s", self)
        self._inventory = Inventory(self._res_get)

        logger.debug("%s, starting workers", self)

        td = threading.Thread(
            target=self.__class__._handle_workers, args=(self,)
        )
        td.daemon = True
        td._state = RUN
        # `td` will be started later by `lazy_start_workers` later
        self._worker_handler = td
        self._worker_handler_started = False

        # launch task handler
        td = threading.Thread(
            target=self._handle_tasks,
        )
        td.daemon = True
        td._state = RUN
        td.start()
        self._task_handler = td

    def __repr__(self):
        return "<{}({}, {})>".format(
            type(self).__name__,
            getattr(self, "_processes", None),
            getattr(self, "_master_addr", None),
        )

    def _handle_tasks(self):
        taskq = self.taskq
        master_sock = self._master_sock

        while True:
            if self.sent_tasks - self.recv_tasks > self.max_processing_tasks:
                time.sleep(0.2)
                continue
            task = taskq.get()
            data = pickle.dumps(task)
            master_sock.send(data)
            self.sent_tasks += 1

    def _task_put(self, task):
        self.taskq.put(task)

    def _res_get(self):
        payload = self._result_sock.recv()
        self.recv_tasks += 1
        data = pickle.loads(payload)

        return data

    @staticmethod
    def _join_exited_workers(workers):

        thread = threading.current_thread()

        logger.debug("ZPool _join_exited_workers running, workers %s, "
                     "thread._state %s", workers, thread._state)
        exited_workers = []

        for i in reversed(range(len(workers))):
            if thread._state != RUN:
                break

            worker = workers[i]
            if worker._start_failed:
                exited_workers.append(worker)
                del workers[i]
                logger.debug("ZPool _join_exited_workers running, "
                             "worker._start_failed +1 %s", worker)
                continue

            if worker.exitcode is not None:
                worker.join()
                exited_workers.append(worker)
                del workers[i]
                logger.debug("ZPool _join_exited_workers running, "
                             "worker.join() done +1 %s", worker)
                continue

        logger.debug("ZPool _join_exited_workers finished, workers %s, "
                     "exited %s", workers, exited_workers)

        return exited_workers

    @staticmethod
    def _maintain_workers(processes, workers, master_addr, result_addr, initializer,
                          initargs, maxtasksperchild):
        thread = threading.current_thread()

        workers_per_fp = config.cpu_per_job
        left = processes - len(workers)

        logger.debug("ZPool _maintain_workers running, workers %s", workers)

        threads = []
        while left > 0 and thread._state == RUN:

            if left > workers_per_fp:
                n = workers_per_fp
            else:
                n = left

            master_conn = LazyZConnection(("r", master_addr))
            result_conn = LazyZConnection(("w", result_addr))
            master_conn.set_name("master_conn")
            w = fiber.process.Process(target=zpool_worker,
                                      args=(master_conn,
                                            result_conn,
                                            initializer,
                                            initargs,
                                            maxtasksperchild,
                                            False,
                                            n),
                                      daemon=False)

            w.name = w.name.replace("Process", "PoolWorker")

            td = threading.Thread(target=safe_start, args=(w,))
            td.start()
            threads.append(td)
            logger.debug("started safe_start thread %s", td)

            # w.start()
            logger.debug("started proc %s", w)
            workers.append(w)
            left -= n

        for td in threads:
            logger.debug("joining safe_start thread %s", td)
            td.join(2)
            logger.debug("joining safe_start thread %s finished", td)

        logger.debug("ZPool _maintain_workers finished, workers %s", workers)

    @staticmethod
    def _handle_workers(pool):
        logger.debug("%s _handle_workers running", pool)
        td = threading.current_thread()

        ZPool._maintain_workers(
            pool._processes, pool._pool,
            pool._master_addr, pool._result_addr, pool._initializer,
            pool._initargs, pool._maxtasksperchild
        )

        while td._state == RUN:
            if len(ZPool._join_exited_workers(pool._pool)) > 0:
                # create new workers when old workers exited
                ZPool._maintain_workers(
                    pool._processes, pool._pool,
                    pool._master_addr, pool._result_addr, pool._initializer,
                    pool._initargs, pool._maxtasksperchild
                )

            time.sleep(0.5)

        logger.debug("%s _handle_workers finished. Status is not RUN",
                     pool)

    @staticmethod
    def _chunks(iterable, size):
        for i in range(0, len(iterable), size):
            yield iterable[i:i + size]

    def apply_async(self, func, args=(), kwds={}, callback=None,
                    error_callback=None):
        """
        Run function `func` with arguments `args` and keyword arguments `kwds`
        on a remote Pool worker. This is an asynchronous version of `apply`.

        :param func: target function to run.
        :param args: positional arguments that needs to be passed to `func`.
        :param kwds: keyword arguments that needs to be passed to `func`.
        :param callback: Currently not supported. A callback function that will
            be called when the result is ready.
        :param error_callback: Currently not supported. A callback function
            that will be called when an error occurred.

        :returns: An ApplyResult object which has a method `.get()` to get
            the actual results.
        """

        if self._state != RUN:
            raise ValueError("Pool is not running")
        # assert kwds == {}, 'kwds not supported yet'
        self.lazy_start_workers(func)

        seq = self._inventory.add(1)
        self._task_put((seq, 0, func, [(args, kwds)], True))
        res = ApplyResult(seq, self._inventory)

        return res

    def start_workers(self):
        self._worker_handler.start()
        self._worker_handler_started = True

    def lazy_start_workers(self, func):
        if hasattr(func, "__fiber_meta__"):
            if (
                not hasattr(zpool_worker, "__fiber_meta__")
                or zpool_worker.__fiber_meta__ != func.__fiber_meta__
            ):
                if self._worker_handler_started:
                    raise RuntimeError(
                        "Cannot run function that has different resource "
                        "requirements acceptable by this pool. Try creating a "
                        "different pool for it."
                    )
                zpool_worker.__fiber_meta__ = func.__fiber_meta__

        if not self._worker_handler_started:
            self.start_workers()

    def map_async(self, func, iterable, chunksize=None, callback=None,
                  error_callback=None):
        """
        For each element `e` in `iterable`, run `func(e)`. The workload is
        distributed between all the Pool workers. This is an asynchronous
        version of `map`.


        :param func: target function to run.
        :param iterable: an iterable object to be mapped.
        :param chunksize: if set, elements in `iterable` will be put in to
            chunks whose size is decided by `chunksize`. These chunks will be
            sent to Pool workers instead of each elements in `iterable`. If not
            set, the chunksize is decided automatically.
        :param callback: Currently not supported. A callback function that will
            be called when the result is ready.
        :param error_callback: Currently not supported. A callback function
            that will be called when an error occurred.

        :returns: An MapResult object which has a method `.get()` to get
            the actual results.
        """

        if error_callback:
            # TODO(jiale) implement error callback
            raise NotImplementedError

        if self._state != RUN:
            raise ValueError("Pool is not running")

        if chunksize is None:
            chunksize = 32

        if not hasattr(iterable, '__len__'):
            iterable = list(iterable)

        self.lazy_start_workers(func)

        seq = self._inventory.add(len(iterable))

        chunks = self.__class__._chunks(iterable, chunksize)
        for batch, chunk in enumerate(chunks):
            self._task_put((seq, batch * chunksize, func, chunk, False))
        res = MapResult(seq, self._inventory)

        return res

    def apply(self, func, args=(), kwds={}):
        """
        Run function `func` with arguments `args` and keyword arguments `kwds`
        on a remote Pool worker.

        :param func: target function to run.
        :param args: positional arguments that needs to be passed to `func`.
        :param kwds: keyword arguments that needs to be passed to `func`.

        :returns: the return value of `func(*args, **kwargs)`.
        """
        return self.apply_async(func, args, kwds).get()

    def map(self, func, iterable, chunksize=None):
        """
        For each element `e` in `iterable`, run `func(e)`. The workload is
        distributed between all the Pool workers.


        :param func: target function to run.
        :param iterable: an iterable object to be mapped.
        :param chunksize: if set, elements in `iterable` will be put in to
            chunks whose size is decided by `chunksize`. These chunks will be
            sent to Pool workers instead of each elements in `iterable`. If not
            set, the chunksize is decided automatically.

        :returns: A list of results equivalent to calling
            `[func(x) for x in iterable]`.
        """
        logger.debug('%s map func=%s', self, func)
        return self.map_async(func, iterable, chunksize).get()

    def imap(self, func, iterable, chunksize=1):
        """
        For each element `e` in `iterable`, run `func(e)`. The workload is
        distributed between all the Pool workers. This function returns an
        iterator which user and iterate over to get results.


        :param func: target function to run.
        :param iterable: an iterable object to be mapped.
        :param chunksize: if set, elements in `iterable` will be put in to
            chunks whose size is decided by `chunksize`. These chunks will be
            sent to Pool workers instead of each elements in `iterable`. If not
            set, the chunksize is decided automatically.

        :returns: an iterator which user can use to get results.
        """
        res = self.map_async(func, iterable, chunksize)
        return res.iget_ordered()

    def imap_unordered(self, func, iterable, chunksize=1):
        """
        For each element `e` in `iterable`, run `func(e)`. The workload is
        distributed between all the Pool workers. This function returns an
        **unordered** iterator which user and iterate over to get results.
        This means that the order of the results may not match the order of
        the `iterable`.


        :param func: target function to run.
        :param iterable: an iterable object to be mapped.
        :param chunksize: if set, elements in `iterable` will be put in to
            chunks whose size is decided by `chunksize`. These chunks will be
            sent to Pool workers instead of each elements in `iterable`. If not
            set, the chunksize is decided automatically.

        :returns: an unordered iterator which user can use to get results.
        """
        res = self.map_async(func, iterable, chunksize)
        return res.iget_unordered()

    def starmap_async(self, func, iterable, chunksize=None, callback=None,
                      error_callback=None):
        """
        For each element `args` in `iterable`, run `func(*args)`. The workload
        is distributed between all the Pool workers. This is an asynchronous
        version of `starmap`.

        For example, `starmap_async(func, [(1, 2, 3), (4, 5, 6)])` will result
        in calling `func(1, 2, 3)` and `func(4, 5, 6)` on a remote host.


        :param func: target function to run.
        :param iterable: an iterable object to be mapped.
        :param chunksize: if set, elements in `iterable` will be put in to
            chunks whose size is decided by `chunksize`. These chunks will be
            sent to Pool workers instead of each elements in `iterable`. If not
            set, the chunksize is decided automatically.
        :param callback: Currently not supported. A callback function that will
            be called when the result is ready.
        :param error_callback: Currently not supported. A callback function
            that will be called when an error occurred.

        :returns: An MapResult object which has a method `.get()` to get
            the actual results.
        """

        if self._state != RUN:
            raise ValueError("Pool is not running")

        if chunksize is None:
            chunksize = 32

        if not hasattr(iterable, '__len__'):
            iterable = list(iterable)

        self.lazy_start_workers(func)

        seq = self._inventory.add(len(iterable))

        iterable = [(item,) for item in iterable]

        chunks = self.__class__._chunks(iterable, chunksize)
        for batch, chunk in enumerate(chunks):
            self._task_put((seq, batch * chunksize, func, chunk, True))

        res = MapResult(seq, self._inventory)

        return res

    def starmap(self, func, iterable, chunksize=None):
        """
        For each element `args` in `iterable`, run `func(*args)`. The workload
        is distributed between all the Pool workers.

        For example, `starmap_async(func, [(1, 2, 3), (4, 5, 6)])` will result
        in calling `func(1, 2, 3)` and `func(4, 5, 6)` on a remote host.


        :param func: target function to run.
        :param iterable: an iterable object to be mapped.
        :param chunksize: if set, elements in `iterable` will be put in to
            chunks whose size is decided by `chunksize`. These chunks will be
            sent to Pool workers instead of each elements in `iterable`. If not
            set, the chunksize is decided automatically.
        :param callback: Currently not supported. A callback function that will
            be called when the result is ready.
        :param error_callback: Currently not supported. A callback function
            that will be called when an error occurred.

        :returns: A list of results equivalent to calling
            `[func(*arg) for arg in iterable]`
        """
        return self.starmap_async(func, iterable, chunksize).get()

    def _send_sentinels_to_workers(self):
        logger.debug('send sentinels(None) to workers %s', self)
        for i in range(self._processes):
            self._task_put(None)

    def close(self):
        """
        Close this Pool. This means the current pool will be put in to a
        closing state and it will not accept new tasks. Existing workers will
        continue to work on tasks that have been dispatched to them and exit
        when all the tasks are done.
        """
        logger.debug('closing pool %s', self)
        if self._state == RUN:
            self._state = CLOSE
            self._worker_handler._state = CLOSE

            for p in self._pool:
                if hasattr(p, '_sentinel'):
                    p._state = CLOSE

            self._send_sentinels_to_workers()

    def terminate(self):
        """
        Terminate this pool. This means that this pool will be terminated and
        all its pool workers will also be terminated. Task that have been
        dispatched will be discarded.
        """
        logger.debug('terminating pool %s', self)

        logger.debug('set pool._worker_handler.status = TERMINATE')
        self._worker_handler._state = TERMINATE
        self._state = TERMINATE

        for p in self._pool:
            p._state = TERMINATE

        pool = self._pool
        N = min(100, len(pool))
        # Thread pool used for process termination. With randomized delay, this
        # roughly equals N requests per second.
        tpool = ThreadPool(N)

        # Terminate workers which haven't already finished.
        if pool and hasattr(pool[0], 'terminate'):
            logger.debug('terminating workers')
            # tpool.map_async(safe_terminate_worker, pool)
            tpool.map(safe_terminate_worker, pool)

        if pool and hasattr(pool[0], 'terminate'):
            logger.debug("joining pool workers, this may take some "
                         "time. workers: %s", pool)
            tpool.map(safe_join_worker, pool)

        logger.debug("joining pool._worker_handler")
        self._worker_handler.join()

    def join(self):
        """
        Wait for all the pool workers of this pool to exit. This should be
        used after `terminate()` or `close()` are called on this pool.
        """
        logger.debug('%s.join()', self)
        assert self._state in (TERMINATE, CLOSE)

        for p in self._pool:
            if p._state not in (TERMINATE, CLOSE):
                logger.debug("%s.join() ignore newly connected Process %s",
                             self, p)
                continue
            p.join()

    def wait_until_workers_up(self):
        logger.debug('%s begin wait_until_workers_up', self)
        workers_per_fp = config.cpu_per_job
        n = math.ceil(float(self._processes) / workers_per_fp)

        while len(self._pool) < n:
            logger.debug("%s waiting for all workers to be up, expected %s, "
                         "got %s", self, n, self._pool)
            time.sleep(0.5)

        for p in self._pool:
            logger.debug('%s waiting for _sentinel %s', self, p)
            while not hasattr(p, '_sentinel') or p._sentinel is None:
                time.sleep(0.5)
        # now all the worker has connected to the master, wait
        # for some additional time to be sure.
        time.sleep(1)
        logger.debug('%s wait_until_workers_up done', self)


class ResilientZPool(ZPool):
    """
    ZPool with error handling. The differences are:

    * Master socket is a ROUTER socket instead of DEALER socket.
    * Add pending table.
    * When an died worker is detected, it's jobs are resubmitted to work Q
      in addition to restarting that worker.

    The API of `ResilientZPool` is the same as `ZPool`. One difference is that
    if `processes` argument is not set, its default value is 1.
    """
    def __init__(self, processes=None, initializer=None, initargs=(),
                 maxtasksperchild=None, cluster=None):

        self.active_peer_dict = {}
        self.active_peer_list = []
        self.peer_lock = threading.Lock()
        self.taskq = queue.Queue()
        self._pending_table = {}

        super(ResilientZPool, self).__init__(
            processes=processes,
            initializer=initializer,
            initargs=initargs,
            maxtasksperchild=maxtasksperchild,
            cluster=cluster,
            master_sock_type="rep")

        '''
        # launch task handler
        td = threading.Thread(
            target=self._handle_tasks,
            args=(
                self.taskq,
                self._master_sock,
                self.active_peer_list,
                self._pending_table,
            ))
        td.daemon = True
        td._state = RUN
        td.start()
        self._task_handler = td
        '''

        self._pid_to_rid = {}

    def _add_peer(self, ident):
        self.peer_lock.acquire()

        self.active_peer_dict[ident] = True
        self._pending_table[ident] = {}
        self.active_peer_list.append(ident)

        self.peer_lock.release()

    def _remove_peer(self, ident):
        # _pendint_table will be cleared later in error handling phase
        self.peer_lock.acquire()

        del self.active_peer_dict[ident]
        self.active_peer_list.remove(ident)

        self.peer_lock.release()

    def _res_get(self):
        # check for system messages
        payload = self._result_sock.recv()
        data = pickle.loads(payload)

        # remove item from pending table
        seq, batch, i, result, ident = data

        # seq, batch, func, chunks, is_starmap
        task = self._pending_table[ident][(seq, batch)]
        chunksize = len(task[3])

        if i == batch + chunksize - 1:
            # this batch has been processed, remove this item from pending
            # table
            del self._pending_table[ident][(seq, batch)]

        # skip ident
        return data[:-1]

    def _task_put(self, task):
        self.taskq.put(task)

    def _handle_tasks(self):
        thread = threading.current_thread()
        taskq = self.taskq
        master_sock = self._master_sock
        pending_table = self._pending_table

        while thread._state == RUN:
            task = taskq.get()

            if task is None:
                logger.debug("_handle_tasks got sentinel")
                break

            msg = master_sock.recv()
            ident, pid = struct.unpack("4si", msg)

            # add peer
            if not self.active_peer_dict.get(ident, None):
                logger.debug('ResilientZPool got REG %s', msg)
                self._add_peer(ident)
                self._pid_to_rid[pid] = ident

            data = pickle.dumps(task)

            seq, batch, func, arg_list, starmap = task
            # use (seq, batch) to identify a (chunked) task. batch is the start
            # number of that batch.
            pending_table[ident][(seq, batch)] = task

            master_sock.send(data)

        # tell peers to exit
        data = pickle.dumps(None)
        #print("send sentinals to workers, active peers", self.active_peer_list)
        for i in range(len(self._pool)):
            #print("waiting for requests")
            msg = master_sock.recv()
            #print("got ", msg)
            master_sock.send(data)
            #print("send ", data)

        #print("exiting handle_tasks ")
        logger.debug('ResilientZPool _handle_tasks exited')

    @staticmethod
    def _maintain_workers(processes, workers, master_addr, result_addr, initializer,
                          initargs, maxtasksperchild):
        thread = threading.current_thread()

        workers_per_fp = config.cpu_per_job
        left = processes - len(workers)

        logger.debug("ResilientZPool _maintain_workers running, workers %s",
                     workers)

        threads = []
        while left > 0 and thread._state == RUN:

            if left > workers_per_fp:
                n = workers_per_fp
            else:
                n = left

            master_conn = LazyZConnection(("req", master_addr))
            result_conn = LazyZConnection(("w", result_addr))

            #conn = LazyZConnectionReg(("req", master_addr))
            master_conn.set_name("master_conn")
            w = fiber.process.Process(target=zpool_worker,
                                      args=(master_conn,
                                            result_conn,
                                            initializer,
                                            initargs,
                                            maxtasksperchild,
                                            False,
                                            n,
                                            True),
                                      daemon=False)

            w.name = w.name.replace("Process", "PoolWorker")

            td = threading.Thread(target=safe_start, args=(w,))
            td.start()
            threads.append(td)
            logger.debug("started safe_start thread %s", td)

            # w.start()
            logger.debug("started proc %s", w)
            workers.append(w)
            left -= n

        for td in threads:
            logger.debug("joining safe_start thread %s", td)
            td.join(2)
            logger.debug("joining safe_start thread %s finished", td)

        logger.debug("ResilientZPool _maintain_workers finished, workers %s",
                     workers)

    @staticmethod
    def _handle_workers(pool):
        logger.debug("%s _handle_workers running", pool)
        td = threading.current_thread()

        ResilientZPool._maintain_workers(
            pool._processes, pool._pool,
            pool._master_addr, pool._result_addr, pool._initializer,
            pool._initargs, pool._maxtasksperchild
        )

        while td._state == RUN:
            exited_workers = ResilientZPool._join_exited_workers(pool._pool)
            if len(exited_workers) > 0:
                # create new workers when old workers exited
                logger.debug("Exited workers %s", exited_workers)

                ResilientZPool._maintain_workers(
                    pool._processes, pool._pool,
                    pool._master_addr, pool._result_addr, pool._initializer,
                    pool._initargs, pool._maxtasksperchild
                )

                # resubmit tasks

                logger.debug("Resubmitting tasks from failed workers")
                for worker in exited_workers:
                    rid = pool._pid_to_rid[worker.pid]
                    # remove rid from active peers
                    pool._remove_peer(rid)

                    # Take care of pending tasks
                    tasks = pool._pending_table[rid]

                    logger.debug("Failed worker %s tasks %s", worker, tasks)
                    for _, task in tasks.items():
                        # resubmit each task
                        pool._task_put(task)
                        logger.debug("Resubmit task %s", task)

                    # Remove tasks from pending table
                    del pool._pending_table[rid]
                    logger.debug("Remove rid %s from pending table", rid)

            time.sleep(0.5)

        logger.debug("%s _handle_workers finished. Status is not RUN",
                     pool)

    def terminate(self):
        self._task_handler._state = TERMINATE

        super(ResilientZPool, self).terminate()

    def close(self):
        logger.debug('closing pool %s', self)
        if self._state == RUN:
            self._state = CLOSE
            self._worker_handler._state = CLOSE

            for p in self._pool:
                if hasattr(p, '_sentinel'):
                    p._state = CLOSE

            #self._send_sentinels_to_workers()

        #logger.debug("ResilientZPool _send_sentinels_to_workers: "
        #             "send to task handler")
        self._task_put(None)
        self._task_handler._state = CLOSE

    def _send_sentinels_to_workers(self):
        logger.debug("ResilientZPool _send_sentinels_to_workers: "
                     "send to workers")
        data = pickle.dumps(None)
        for ident in self.active_peer_list:
            self._master_sock.send_multipart([ident, b"", data])


#Pool = ZPool
Pool = ResilientZPool



import fiber
import multiprocessing as mp
import time
import pytest
import docker
import fiber.config
from fiber.docker_backend import Backend
from fiber.util import find_ip_by_net_interface
import psutil
import ipaddress
import socket


def wrap_result(queue):
    backend = Backend()
    r = backend.get_listen_addr()
    queue.put(r)

delay_secs=iter([0, 1, 2, 3])
class DelayedBackend(Backend):
    def create_job(self, job_spec):
        time.sleep(next(delay_secs))
        return super(DelayedBackend, self).create_job(job_spec)


def square_worker(a):
    return a**2


@pytest.fixture(scope="module")
def client():
    return docker.from_env()


def get_ifce(target_ifce):
    ifces = psutil.net_if_addrs()
    ifce = ifces[target_ifce]

    for sincaddr in ifce:
        if sincaddr.family == socket.AF_INET:
            return sincaddr

    return None

@pytest.mark.skipif(fiber.config.default_backend != "docker",
                    reason="skipped because current backend is not docker")
class TestDockerBackend():
    # See https://docs.pytest.org/en/latest/fixture.html#autouse-fixtures-xunit-setup-on-steroids
    @pytest.fixture(autouse=True)
    def check_leak(self, request, client):
        assert fiber.active_children() == []
        pre_workers = client.containers.list()
        yield
        post_workers = client.containers.list()
        assert pre_workers == post_workers


    def test_get_listen_addr_fp(self):
        ifce = get_ifce("docker0")
        network = ipaddress.IPv4Network((ifce.address, ifce.netmask), strict=False)

        q = fiber.SimpleQueue()
        p = fiber.Process(target=wrap_result, args=(q, ), name="test_get_listen_addr")
        p.start()
        res = q.get()
        ip = res[0]
        assert ipaddress.IPv4Address(ip) in network

        p.join()
        assert p.exitcode == 0

    def test_get_listen_addr_mp(self):
        ifce = get_ifce("docker0")
        network = ipaddress.IPv4Network((ifce.address, ifce.netmask), strict=False)

        q = mp.SimpleQueue()
        p = mp.Process(target=wrap_result, args=(q, ), name="test_get_listen_addr")
        p.start()
        p.join()
        assert p.exitcode == 0
        res = q.get()

        ip = res[0]
        assert ipaddress.IPv4Address(ip) in network

    def test_job_creation_with_delay(self):
        fiber.backend.get_backend(name="docker")
        old_backend = fiber.backend._backends["docker"]
        fiber.backend._backends["docker"] = DelayedBackend()

        p = fiber.Pool(4)
        # explicitly start workers instead of lazy start
        p.start_workers()
        p.wait_until_workers_up()

        res = p.map(square_worker, [1, 2, 3, 4])

        p.terminate()
        fiber.backend._backends["docker"] = old_backend

        assert res == [i**2 for i in range(1, 5)]

        # wait for 2 seconds to let docker finish starting
        #time.sleep(2)
        p.join()



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import multiprocessing as mp
import os
import sys
import threading
import time
import uuid
from os.path import expanduser

import docker
import psutil
import requests

import fiber
import fiber.core as core
import fiber.config as config
from fiber.core import ProcessStatus
from fiber.util import find_ip_by_net_interface, find_listen_address

logger = logging.getLogger('fiber')


STATUS_MAP = {
    "restarting": ProcessStatus.INITIAL,
    "created": ProcessStatus.INITIAL,
    "running": ProcessStatus.STARTED,
    "paused": ProcessStatus.STARTED,
    "exited": ProcessStatus.STOPPED,
}

HOME_DIR = expanduser("~")


class DockerJob(core.Job):
    def update(self):
        container = self.data
        self.host = container.attrs['NetworkSettings']['IPAddress']


class Backend(core.Backend):
    name = "docker"

    def __init__(self):
        # Based on this link, no lock is needed accessing self.client
        # https://github.com/docker/docker-py/issues/619
        self.client = docker.from_env()

    def create_job(self, job_spec):
        logger.debug("[docker]create_job: %s", job_spec)
        cwd = os.getcwd()
        volumes = {cwd: {'bind': cwd, 'mode': 'rw'},
                   HOME_DIR: {'bind': HOME_DIR, 'mode': 'rw'}}
        debug = config.debug

        tty = debug
        stdin_open = debug

        if job_spec.image is None:
            image = config.default_image
        else:
            image = job_spec.image

        try:
            container = self.client.containers.run(
                image,
                job_spec.command,
                name=job_spec.name + '-' + str(uuid.uuid4()),
                volumes=volumes,
                cap_add=["SYS_PTRACE"],
                tty=tty,
                stdin_open=stdin_open,
                detach=True
            )
        except docker.errors.ImageNotFound:
            raise mp.ProcessError(
                "Docker image \"{}\" not found or cannot be pulled from "
                "docker registry.".format(job_spec.image))

        except docker.errors.APIError as e:
            raise mp.ProcessError(
                    "Failed to start docker container: {}".format(e)
            )

        job = DockerJob(container, container.id)
        # delayed
        container._fiber_backend_reloading = False
        return job

    def _reload(self, container):
        container._fiber_backend_reloading = True
        logger.debug("container reloading %s", container.name)
        container.reload()
        if config.debug:
            time.sleep(0.1)
        else:
            time.sleep(1)

        container._fiber_backend_reloading = False

    def get_job_logs(self, job):
        container = job.data
        return container.logs(stream=False).decode('utf-8')

    def get_job_status(self, job):
        container = job.data

        if config.merge_output:
            print(container.logs(stream=False).decode('utf-8'))

        logging.debug("get_job_status: %s", container.status)
        status = STATUS_MAP[container.status]

        if not container._fiber_backend_reloading:
            td = threading.Thread(target=self._reload, args=(container,))
            td.start()
            logger.debug("start container reloading thread %s", container.name)
        return status

    def wait_for_job(self, job, timeout):
        container = job.data
        logger.debug("wait_for_job: %s", container.name)

        if config.merge_output:
            print(container.logs(stream=False).decode('utf-8'))

        # Can't set timeout to 0 for container.wait. Use a very
        # short wait time instead
        if timeout is not None and timeout <= 0:
            timeout = 0.05
        try:
            res = container.wait(timeout=timeout)
        except requests.exceptions.ConnectionError:
            # Docker SDK for Python document said that wait will raise
            # requests.exceptions.ReadTimeout exception, but actually it
            # raises requests.exceptions.ConnectionError.
            # See https://docker-py.readthedocs.io/en/stable/containers.html#docker.models.containers.Container.wait # noqa: E501

            logger.debug("wait_for_job done(ConnectionError): %s",
                         container.name)
            return None
        except requests.exceptions.ReadTimeout:
            # Incase it raises ReadTimeout
            logger.debug("wait_for_job done(timeout): %s", container.name)
            return None

        if config.merge_output:
            print(container.logs(stream=False).decode('utf-8'))

        logger.debug("wait_for_job done: %s", container.name)

        return res['StatusCode']

    def terminate_job(self, job):
        logging.debug("terminate_job")
        container = job.data

        if config.merge_output:
            print(container.logs(stream=False).decode('utf-8'))

        logger.debug("before terminating a container, status: %s",
                     container.status)
        try:
            container.kill()
        except docker.errors.APIError as e:
            if e.status_code == 409:
                logger.warning("container is not running %s", str(e))
                # Ignore 409 error which usually means container is not running
                return
            raise e
        logger.debug("terminate job finished, %s", container.status)

    def get_listen_addr(self):
        ip = None

        if sys.platform == "darwin":
            # use the same hostname for both master and non master process
            # because docker.for.mac.localhost resolves to different inside
            # and outside docker container. "docker.for.mac.localhost" is
            # the name that doesn't change in and outside the container.
            return "docker.for.mac.localhost", 0, 0

        if not isinstance(fiber.current_process(), fiber.Process):
            # not inside docker
            ifce = "docker0"
            ip = find_ip_by_net_interface(ifce)
        else:
            # inside a Fiber process
            ip, ifce = find_listen_address()

        if ip is None:
            raise mp.ProcessError(
                "Can't find a usable IPv4 address to listen. ifce_name: {}, "
                "ifces: {}".format(ifce, psutil.net_if_addrs()))
        # use 0 to bind to a random free port number
        return ip, 0, ifce



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import fiber.config as config
from fiber import process


class FiberContext():
    _name = ''
    Process = process.Process

    current_process = staticmethod(process.current_process)
    active_children = staticmethod(process.active_children)

    def Manager(self):
        """Returns a manager associated with a running server process

        The managers methods such as `Lock()`, `Condition()` and `Queue()`
        can be used to create shared objects.
        """
        from fiber.managers import SyncManager
        m = SyncManager()
        m.start()
        return m

    def Pool(self, processes=None, initializer=None, initargs=(),
             maxtasksperchild=None, error_handling=False):
        """Returns a process pool object"""
        from .pool import ZPool, ResilientZPool
        if error_handling:
            return ResilientZPool(processes, initializer, initargs, maxtasksperchild)
        else:
            return ZPool(processes, initializer, initargs, maxtasksperchild)

    def SimpleQueue(self):
        """Returns a queue object"""
        if config.use_push_queue:
            from .queues import SimpleQueuePush
            return SimpleQueuePush()

        # PullQueue is not supported anymore
        raise NotImplementedError

    def Pipe(self, duplex=True):
        """Returns two connection object connected by a pipe"""
        from .queues import Pipe
        return Pipe(duplex)

    def cpu_count(self):
        return os.cpu_count()

    def get_context(self, method=None):
        if method is None:
            return self
        if method != "spawn":
            raise ValueError("Fiber only supports spawn context")
        return _concrete_contexts[method]


_concrete_contexts = {
        'spawn': FiberContext()
}

_default_context = _concrete_contexts['spawn']



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import fcntl
import io
import itertools
import logging
import multiprocessing as mp
import os
import socket
import struct
import subprocess
import sys
import threading
import time
from multiprocessing import reduction, spawn, util
from multiprocessing.context import set_spawning_popen

import cloudpickle
import requests

import fiber.util
import fiber.config as config
from fiber.backend import get_backend
from fiber.core import JobSpec
from fiber.core import ProcessStatus

logger = logging.getLogger("fiber")

# each line much ends with `;`
fiber_init_start = """
import sys;
import os;
import socket;
import struct;
os.environ["FIBER_WORKER"]="1";
os.chdir("{cwd}");
import fiber;
import fiber.spawn;
from multiprocessing import spawn, reduction;
sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM);
"""

fiber_init_net_active = """
sock.connect(("{host}", {port}));
conn = sock;
"""

fiber_init_net_passive = """
sock.bind(('', {port}));
sock.listen(0);
conn, _ = sock.accept();
"""

fiber_init_end = """
conn.send(struct.pack("<I", {id}));
fd = conn.fileno();
exitcode = fiber.spawn.spawn_prepare(fd);
sys.exit(exitcode)
"""

fiber_init_start = fiber_init_start.strip()
fiber_init_net_active = fiber_init_net_active.strip()
fiber_init_net_passive = fiber_init_net_passive.strip()
fiber_init_end = fiber_init_end.strip()

admin_host = None
admin_port = None
_fiber_background_thread = None
_fiber_background_thread_lock = threading.Lock()
_event_dict = {}
_event_counter = itertools.count(1)


def get_fiber_init():
    if config.ipc_active:
        fiber_init = fiber_init_start + fiber_init_net_active + fiber_init_end
    else:
        fiber_init = fiber_init_start + fiber_init_net_passive + fiber_init_end

    fiber_init = fiber_init.replace("\n", "").strip()
    return fiber_init


def fiber_background(listen_addr, event_dict):
    global admin_host, admin_port

    # Background thread for handling inter fiber process admin traffic
    ip, port = listen_addr
    host = ""

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((host, port))
    logger.debug("fiber background thread attempt to bind %s:%s", host, port)
    # fix port number (could be 0 previously)
    _, port = sock.getsockname()
    logger.debug("fiber background thread bind addr is to %s:%s", host, port)
    # backlog argument becomes optional since Python 3.5, but in order
    # to make Fiber work with 3.4, here we set a reasonable listen backlog
    # size. See https://docs.python.org/3.4/library/socket.html
    sock.listen(5)

    admin_host = ip
    admin_port = port

    sentinel = event_dict[-1]
    # notifi master thread that background thread is ready
    sentinel.set()
    logger.debug("fiber_background thread ready")
    while True:
        conn, addr = sock.accept()
        logger.debug("got connection from %s", addr)
        # fixed 4 byte id
        buf = conn.recv(4)
        # struct.unpack returns a tuple event if the result only has
        # one element
        ident = struct.unpack("<I", buf)[0]
        event = event_dict.get(ident, None)
        if event is None:
            logger.warn(
                "something is wrong, no event found for this id: %s", ident
            )
            continue
        logger.debug("got connection for id %s", ident)
        event_dict[ident] = conn
        event.set()


def get_python_exe(backend_name):
    if backend_name == "docker":
        # TODO(jiale) fix python path
        python_exe = "/usr/local/bin/python"
    else:
        python_exe = sys.executable

    logger.debug("backend is \"%s\", use python exe \"%s\"",
                 backend_name, python_exe)
    return python_exe


def get_pid_from_jid(jid):
    # Some Linux system has 32768 as max pid number. 32749 is a prime number
    # close to 32768.
    return hash(jid) % 32749


class Popen(object):
    method = "spawn"

    def __del__(self):
        if getattr(self, "ident", None):
            # clean up entry in event_dict
            global _event_dict
            #logger.debug("cleanup entry _event_dict[%s]", self.ident)
            _event_dict.pop(self.ident, None)

    def __repr__(self):
        return "<{}({})>".format(
            type(self).__name__, getattr(self, "process_obj", None)
        )

    def __init__(self, process_obj, backend=None, launch=False):
        self.returncode = None
        self.backend = get_backend()

        ip, _, _ = self.backend.get_listen_addr()

        self.master_host = ip
        self.master_port = config.ipc_admin_master_port
        self.worker_port = config.ipc_admin_worker_port

        self.sock = None
        self.host = ""

        self.job = None
        self.pid = None
        self.process_obj = process_obj
        self._exiting = None
        self.sentinel = None
        self.ident = None

        if launch:
            self._launch(process_obj)

    def launch_fiber_background_thread_if_needed(self):
        global _fiber_background_thread_lock
        _fiber_background_thread_lock.acquire()

        global _fiber_background_thread
        if _fiber_background_thread is not None:
            _fiber_background_thread_lock.release()
            return

        try:
            logger.debug(
                "_fiber_background_thread is None, creating "
                "background thread"
            )
            # Create a background thread to handle incoming connections
            # from fiber child processes
            event = threading.Event()
            event.clear()
            _event_dict[-1] = event
            td = threading.Thread(
                target=fiber_background,
                args=((self.master_host, self.master_port), _event_dict),
                daemon=True,
            )
            td.start()
        except Exception as e:
            raise e
        finally:
            logger.debug("waiting for background thread")
            event.wait()
            logger.debug(
                "master received message that fiber_background thread is ready"
            )
            _fiber_background_thread = td
            _fiber_background_thread_lock.release()

    def get_command_line(self, **kwds):
        """Returns prefix of command line used for spawning a child process."""
        prog = get_fiber_init()
        prog = prog.format(**kwds)
        opts = util._args_from_interpreter_flags()
        if config.debug:
            return (
                [get_python_exe(self.backend.name)]
                + opts
                + ["-u", "-c", prog, "--multiprocessing-fork"]
            )

        return (
            [get_python_exe(self.backend.name)]
            + opts
            + ["-c", prog, "--multiprocessing-fork"]
        )

    def _accept(self):
        conn, addr = self.sock.accept()
        logger.debug("successfully accept")
        # TODO verify if it's the same client
        return conn

    def _get_job(self, cmd):
        spec = JobSpec(
            command=cmd,
            image=config.image,
            name=self.process_obj.name,
            cpu=config.cpu_per_job,
            mem=config.mem_per_job,
        )
        if hasattr(self.process_obj._target, "__self__"):
            metadata = getattr(
                self.process_obj._target.__self__, "__fiber_meta__", None
            )
        else:
            metadata = getattr(self.process_obj._target, "__fiber_meta__", None)
        if metadata:
            for k, v in metadata.items():
                setattr(spec, k, v)

        return spec

    def _run_job(self, job):
        try:
            job = self.backend.create_job(job)
        except requests.exceptions.ReadTimeout as e:
            raise mp.TimeoutError(str(e))
        self.job = job

        return job

    def _sentinel_readable(self, timeout=0):
        # Use fcntl(fd, F_GETFD) instead of select.* becuase:
        # * select.select() can't work with fd > 1024
        # * select.poll() is not thread safe
        # * select.epoll() is Linux only
        # Also, fcntl(fd, F_GETFD) is cheaper than the above calls.
        return fcntl.fcntl(self.sentinel, fcntl.F_GETFD)

    def poll(self, flag=os.WNOHANG):

        # returns None if the process is not stopped yet. Otherwise, returns
        # process exit code.
        if flag != os.WNOHANG:
            raise NotImplementedError("flag {} is not supported".format(flag))

        # no need to wait here. get_job_status will be called later
        # if self._exiting:
        #     logger.debug("[poll] self._exiting is True, calling self.wait.")
        #     return self.wait(timeout=1)

        if self.job is None:
            # self.job is None meaning this process hasn't been fully started
            # yet.
            # logger.debug("[poll] self.job is None, return.")
            return None

        if self.returncode is None:
            if self.sentinel is not None:
                if not self._sentinel_readable(0):
                    logger.debug(
                        "self.sentinel not readable, process is still up"
                    )

                    # process is still up
                    return None
                logger.debug("self.sentinel readable, process is probably down")

            logger.debug("self.sentinel is None, getting status")
            status = self.backend.get_job_status(self.job)
            # When exiting, make sure to call wait to get exit code
            if status != ProcessStatus.STOPPED and (not self._exiting):
                return None
        return self.wait(timeout=0)

    def wait(self, timeout=None):
        if self.job is None:
            # self.job is None meaning this process hasn't been fully started
            # yet.
            # TODO(jiale) timeout is ignored here, but since the job hasn't
            # been started yet, there is nothing to wait. This can happen when
            # Process.start() is called in another thread and usually it takes
            # a long time to start a new Process, so self.job can be empty.
            return None

        if self.returncode is None:
            code = self.backend.wait_for_job(self.job, timeout)
            if code is None:
                return code
            assert type(code) is int
            self.returncode = code
        return self.returncode

    def _pickle_data(self, data, fp):
        if fiber.util.is_in_interactive_console():
            logger.debug("in interactive shell, use cloudpickle")
            cloudpickle.dump(data, fp)
        else:
            logger.debug("not in interactive shell, use reduction")
            reduction.dump(data, fp)

    def _launch(self, process_obj):
        logger.debug("%s %s _launch called", process_obj, self)

        if config.ipc_active:
            logger.debug("%s ipc_active is set, launch background thread", self)
            self.launch_fiber_background_thread_if_needed()
        else:
            logger.debug("%s ipc_active is not set", self)

        # Setup networking
        ident = next(_event_counter)
        self.ident = ident
        # this needs to happen after self._setup_listen where port is decided
        global admin_host, admin_port
        if config.ipc_active:
            # in active mode, port is admin_port which could be 0 at first and
            # set to the actual port by the backend thread
            port = admin_port
        else:
            # in passive mode, port can't be 0 because this need to be
            # pre-determined information between master and workers.
            assert self.worker_port != 0, (
                "port can't be 0 because this "
                "need to be pre-determined information between master "
                "and workers."
            )
            port = self.worker_port

        if config.ipc_active:
            assert admin_host is not None
            assert admin_port is not None

        cmd = self.get_command_line(
            cwd=os.getcwd(), host=admin_host, port=port, id=ident
        )

        job = self._get_job(cmd)

        event = threading.Event()
        event.clear()
        _event_dict[ident] = event
        logger.debug(
            "%s popen_fiber_spawn created event %s and set _event_dict[%s]",
            self,
            event,
            ident,
        )

        # prepare data and serialize
        prep_data = spawn.get_preparation_data(process_obj.name)
        prep_data["fiber_config"] = config.get_object()
        # TODO(jiale) what is a better way to setup sys_path inside containers?
        prep_data.pop("sys_path", None)
        logger.debug("%s prep_data: %s", self, str(prep_data))
        fp = io.BytesIO()
        set_spawning_popen(self)
        try:
            self._pickle_data(prep_data, fp)
            self._pickle_data(process_obj, fp)
        finally:
            set_spawning_popen(None)

        # Set process_obj._popen = self here so that we have have the
        # Process <-> Popen binding before _run_job() is called. After
        # _run_job() is called, we started a job with Fiber backend. So
        # Process <-> Popen has to exist so that Process.terminate() can
        # find Popen object and terminate the underlying job. Also, when
        # we serize process_obj above, we have to make sure that
        # Process <-> Popen binding doesn't exist so that Popen doesn't get
        # serialized becuase 1) it's not necessary and 2) some part of Popen
        # cannot be pickled.
        process_obj._popen = self

        # launch job
        job = self._run_job(job)
        self.pid = get_pid_from_jid(job.jid)
        # Fix process obj's pid
        process_obj.ident = self.pid
        post_data = {"pid": self.pid}
        self._pickle_data(post_data, fp)

        send_buffer = fp.getbuffer()

        if config.ipc_active:
            # (worker) active mode, wait for event to be set by
            # background thread
            done = False
            while not done:
                if self._exiting:
                    logger.debug(
                        "process is exiting, don't wait for job to "
                        "connect back"
                    )
                    return
                done = event.wait(0.5)
                status = self.check_status()
                if status == ProcessStatus.STOPPED:
                    return

            logger.debug(
                "popen_fiber_spawn is waiting for accept event %s to finish",
                event,
            )
            conn = _event_dict[ident]
            logger.debug("got conn from _event_counter[%s]", ident)
            del _event_dict[ident]
            logger.debug("remove entry _event_counter[%s]", ident)
        else:
            # (worker) passive mode, check job status until it's ready
            while True:
                if self._exiting:
                    logger.debug(
                        "process is exiting, don't wait for job to start"
                    )
                    return
                status = self.check_status()
                if status == ProcessStatus.STARTED:
                    break
                logger.debug(
                    "waiting 1s for job to start, current status: %s", status
                )
                time.sleep(1)
            # connect to woker job
            if job.host is None:
                job.update()
            ip = job.host
            port = self.worker_port
            addr = (ip, port)

            conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            while True:
                if self._exiting:
                    logger.debug(
                        "process is exiting, don't try to connect to the job"
                    )
                    return
                try:
                    logger.debug("connecting to %s", addr)
                    conn.connect(addr)
                except ConnectionRefusedError:
                    # not ready yet, sleep
                    logger.debug(
                        "Fiber worker is not up yet, waiting 1s for the "
                        "worker to listen on admin address: %s",
                        addr,
                    )
                    time.sleep(1)
                    continue
                break

        logger.debug("send buffer")
        conn.send(send_buffer)

        # logger.debug("sent fp.getbuffer() to sub process")

        self.sentinel = conn
        logger.debug("_launch finished")

    def check_status(self):
        status = self.backend.get_job_status(self.job)
        if status == ProcessStatus.STOPPED:
            # something happened that caused Fiber process to hit an early stop
            msg = self.backend.get_job_logs(self.job)
            logger.error(
                "Failed to start Fiber process %s. Logs from inside the "
                "container:\n%s",
                self.process_obj.name,
                msg,
            )

        return status

    def terminate(self):
        logger.debug("[Popen]terminate() called")

        self._exiting = True
        if self.job is not None:
            logger.debug(
                "[Popen]terminate():self.job is not None, terminate job"
            )
            self.backend.terminate_job(self.job)
        else:
            logger.error("[Popen]terminate():self.job is None, terminate job")
            # raise RuntimeError("[Popen]terminate():self.job is "
            #                    "None %s", self.process_obj)



#
# Module providing various facilities to other parts of the package
#
# multiprocessing/util.py
#
# Copyright (c) 2006-2008, R Oudkerk
# Licensed to PSF under a Contributor Agreement.
#
# Modifications Copyright (c) 2020 Uber Technologies

import itertools
import logging
import multiprocessing.util as mpu
import os
import re
import sys
import socket
import threading
import weakref

import psutil


logger = logging.getLogger('fiber')
_afterfork_registry = weakref.WeakValueDictionary()
_afterfork_counter = itertools.count()


_finalizer_registry = {}
_finalizer_counter = itertools.count()


def register_after_fork(obj, func):
    _afterfork_registry[(next(_afterfork_counter), id(obj), func)] = obj


def _run_after_forkers():
    logging.debug('_fun_after_forkers called')
    items = list(_afterfork_registry.items())
    items.sort()
    for (index, ident, func), obj in items:
        try:
            logging.debug('run after forker %s(%s)', func, obj)
            func(obj)
        except Exception as e:
            logging.info('after forker raised exception %s', e)


class Finalize(mpu.Finalize):
    """Basically this is the same as multiprocessing's Finalize class except
    this one uses it's own _finalizer_registry.
    """
    def __init__(self, obj, callback, args=(), kwargs=None, exitpriority=None):
        assert exitpriority is None or type(exitpriority) is int

        if obj is not None:
            self._weakref = weakref.ref(obj, self)
        else:
            assert exitpriority is not None

        self._callback = callback
        self._args = args
        self._kwargs = kwargs or {}
        self._key = (exitpriority, next(_finalizer_counter))
        self._pid = os.getpid()

        _finalizer_registry[self._key] = self


def find_ip_by_net_interface(target_interface):
    """Returns ip, debug_info."""
    ifces = psutil.net_if_addrs()
    ip = None
    for ifce in ifces:
        # See https://docs.docker.com/v17.09/engine/userguide/networking/default_network/custom-docker0/ # noqa E501
        if ifce == target_interface:
            addrs = ifces[ifce]
            for snicaddr in addrs:
                # IPv4 only
                if snicaddr.family == socket.AF_INET:
                    ip = snicaddr.address
                    return ip
    return None


class ForkAwareThreadLock(object):
    def __init__(self):
        self._reset()
        register_after_fork(self, ForkAwareThreadLock._reset)

    def _reset(self):
        self._lock = threading.Lock()
        self.acquire = self._lock.acquire
        self.release = self._lock.release

    def __enter__(self):
        return self._lock.__enter__()

    def __exit__(self, *args):
        return self._lock.__exit__(*args)


class ForkAwareLocal(threading.local):
    def __init__(self):
        register_after_fork(self, lambda obj: obj.__dict__.clear())

    def __reduce__(self):
        return type(self), ()


def find_listen_address():
    """Find an IP address for Fiber to use."""
    ip = None
    ifce = None
    ifces = psutil.net_if_addrs()
    for ifce, addrs in ifces.items():
        if re.match(r"^eth", ifce) or re.match(r"^en", ifce):
            for snicaddr in addrs:
                # IPv4 only
                if snicaddr.family == socket.AF_INET:
                    ip = snicaddr.address
                    break

    return ip, ifce


def is_in_interactive_console():
    if hasattr(sys, 'ps1'):
        return True

    return False



import fiber
import time
import pytest
import docker
from fiber.managers import AsyncManager


@pytest.fixture(scope="module")
def client():
    return docker.from_env()


def f(ns, ls, di):
    ns.x += 1
    ns.y[0] += 1
    ns_z = ns.z
    ns_z[0] += 1
    ns.z = ns_z

    ls[0] += 1
    ls[1][0] += 1  # unmanaged, not assigned back
    ls_2 = ls[2]   # unmanaged...
    ls_2[0] += 1
    ls[2] = ls_2   # ... but assigned back
    ls[3][0] += 1  # managed, direct manipulation

    di[0] += 1
    di[1][0] += 1  # unmanaged, not assigned back
    di_2 = di[2]   # unmanaged...
    di_2[0] += 1
    di[2] = di_2   # ... but assigned back
    di[3][0] += 1  # managed, direct manipulation


class Calculator():
    def add(self, a, b):
        return a + b

    def add_with_delay(self, a, b, delay=1):
        # delay for 1 seconds
        time.sleep(1)
        return a + b


class MyManager(AsyncManager):
    pass


class TestManagers():

    @pytest.fixture(autouse=True)
    def check_leak(self, request, client):
        assert fiber.active_children() == []
        if fiber.config.default_backend != "docker":
            yield
            return
        pre_workers = client.containers.list()
        yield
        post_workers = client.containers.list()
        assert pre_workers == post_workers

    # This test is adapted from
    # https://stackoverflow.com/questions/9436757/how-does-multiprocessing-manager-work-in-python

    def test_managers_basic(self):
        manager = fiber.Manager()
        ns = manager.Namespace()
        ns.x = 1
        ns.y = [1]
        ns.z = [1]
        ls = manager.list([1, [1], [1], manager.list([1])])
        di = manager.dict({0: 1, 1: [1], 2: [1], 3: manager.list([1])})

        p = fiber.Process(target=f, args=(ns, ls, di), name="test_managers_basic")
        p.start()
        p.join()
        assert ns.x == 2
        assert ns.y == [1]
        assert ns.z == [2]

        assert ls[0] == 2
        assert ls[1] == [1]
        assert ls[2] == [2]
        assert ls[3][0] == 2
        assert len(ls) == 4

        assert di[0] == 2
        assert di[1] == [1]
        assert di[2] == [2]
        assert di[3][0] == 2

    def test_async_manager_basic(self):
        MyManager.register("Calculator", Calculator)

        manager = MyManager()
        manager.start()

        calc = manager.Calculator()

        res = calc.add(10, 32)

        assert res.get() == 42

        total = 4
        managers = [MyManager() for i in range(total)]

        [manager.start() for manager in managers]

        calculators = [manager.Calculator() for manager in managers]

        start_ts = time.time()
        handles = [calc.add_with_delay(10, 32) for calc in calculators]

        results = [handle.get() for handle in handles]

        duration = time.time() - start_ts

        assert duration < 2
        assert results == [42 for i in range(total)]



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


VALID_META_KEYS = ["cpu", "memory", "gpu"]


def post_process(metadata):
    # memory should be in MB
    if "memory" in metadata:
        memory = metadata.pop("memory")
        metadata["mem"] = memory

    return metadata


def meta(**kwargs):
    """
    fiber.meta API allows you to decorate your function and provide some hints to
    Fiber. Currently this is mainly used for specify the resource usage of user
    functions.

    Currently, support keys are:

    | key                   | Type | Default | Notes |
    | --------------------- |:------|:-----|:------|
    | cpu                   | int   | None | The number of CPU cores that this function needs |
    | memory                | int   | None | The size of memory space in MB this function needs |
    | gpu                   | int   | None | The number of GPUs that this function needs. For how to setup Fiber with GPUs, check out [here](advanced.md#using-fiber-with-gpus) |

    Example usage:

    ```python
    @fiber.meta(cpu=4, memory=1000, gpu=1)
    def func():
        do_something()
    ```
    """
    for k in kwargs:
        assert k in VALID_META_KEYS, "Invalid meta argument \"{}\"".format(k)

    def decorator(func):
        meta = post_process(kwargs)
        func.__fiber_meta__ = meta
        return func

    return decorator



# Copyright 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum


MEM_CPU_RATIO = 2  # 2G per cpu


class ProcessStatus(enum.Enum):
    UNKNOWN = 0
    INITIAL = 1
    STARTED = 2
    STOPPED = 3


class JobSpec(object):
    def __init__(self, image=None, command=None, name=None, cpu=None, mem=None,
                 volumes=None, gpu=None):
        # Docker image used to launch this job
        self.image = image
        # Command to run in this job container, this should be a sequence
        # of program arguments instead of a single long string.
        self.command = command
        # Job name
        self.name = name
        # Maximum number of cpu cores this job can use
        self.cpu = cpu
        # Maximum number of cpu cores this job can use
        self.gpu = gpu
        # Maximum memory size in MB that this job can use
        #if mem is None:
        #    mem = cpu * MEM_CPU_RATIO
        self.mem = mem
        # volume name to be mounted, currently only used by k8s backend
        # For example:
        #     volumes = {
        #         "my_volume": {"mode": "rw", "bind": "/persistent"}
        #     }
        self.volumes = volumes

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return '<JobSpec: {}>'.format(vars(self))


class Job(object):
    # Data is used to hold backend specific data associated with this job
    data = None
    # Job id. This is set by backend and should only be used by Fiber backend
    jid = None
    # (Optional) The hostname/IP address for this job, this is used to
    # communicate with the master process. It is only used when
    # `ipc_admin_passive` is enabled.
    host = None

    def __init__(self, data, jid):
        self.data = data
        self.jid = jid

    def update(self):
        # update/refresh job attributes
        raise NotImplementedError


class Backend(object):
    @property
    def name(self):
        raise NotImplementedError

    def create_job(self, job_spec):
        """This function is called when Fiber wants to create a new Process."""
        raise NotImplementedError

    def get_job_status(self, job):
        """This function is called when Fiber wants to to get job status."""
        raise NotImplementedError

    def get_job_logs(self, job):
        """
        This function is called when Fiber wants to to get logs of this job
        """
        return ""

    def wait_for_job(self, job, timeout):
        """Wait for a specific job until timeout. If timeout is None,
        wait until job is done. Returns `None` if timed out or `exitcode`
        if job is finished.
        """
        raise NotImplementedError

    def terminate_job(self, job):
        """Terminate a job described by `job`."""
        raise NotImplementedError

    def get_listen_addr(self):
        """This function is called when Fiber wants to listen on a local
        address for incoming connection. It is currently used by Popen
        and Queue."""
        raise NotImplementedError
